from django.urls import path

from django_advanced_pdf.views.tasks import ViewTaskPDF, PrintReportModal

app_name = 'django_advanced_pdf'

urlpatterns = [
    path('download/<str:file_key>/', ViewTaskPDF.as_view(), name='view_task_pdf'),
    path('download/<str:slug>/modal/', PrintReportModal.as_view(), name='view_task_pdf_modal'),
]



import pathlib
import unittest
from pathlib import Path
import fitz
from reportlab.lib import colors
from reportlab.platypus import TableStyle, Table

from django_advanced_pdf.engine.report_xml import ReportXML
from PIL import Image, ImageChops


class PDFTests(unittest.TestCase):

    def setUp(self):
        pass

    @staticmethod
    def get_test_folder():
        return Path(Path(__file__).resolve().parent, 'test_data')

    def run_report(self, name, object_lookup=None):
        test_folder = self.get_test_folder()
        temp_folder = Path(test_folder, 'temp', name)

        pathlib.Path(temp_folder).mkdir(parents=True, exist_ok=True)

        png_files = list(temp_folder.glob("*.png"))

        # Delete each PNG file
        for png_file in png_files:
            png_file.unlink()

        with open(Path(test_folder, 'reports', f'{name}.xml')) as f:
            xml = f.read()

        report_xml = ReportXML(test_mode=True, object_lookup=object_lookup)
        result = report_xml.load_xml_and_make_pdf(xml=xml)
        matrix = fitz.Matrix(300 / 72, 300 / 72)

        with fitz.open("pdf", result) as doc:
            for page_number, page in enumerate(doc, 1):
                output_file = Path(temp_folder, f'{page_number}.png')

                pix = page.get_pixmap(matrix=matrix)
                pix.pil_save(output_file, format="PNG", dpi=(300, 300))

            page_count = len(doc)

            with open(Path(temp_folder, 'page_count.txt'), 'w') as w:
                w.write(str(page_count))

        self.check_report(name=name, new_page_count=page_count)

    def check_report(self, name, new_page_count):
        test_folder = self.get_test_folder()
        temp_folder = Path(test_folder, 'temp', name)
        held_folder = Path(test_folder, 'held', name)

        with open(Path(held_folder, 'page_count.txt'), 'r') as f:
            self.assertEqual(int(f.read()), new_page_count, msg='Pages count not equal')

        for page_number in range(1, new_page_count+1):

            # Load the two images
            image1 = Image.open(Path(held_folder, f'{page_number}.png'))
            image2 = Image.open(Path(temp_folder, f'{page_number}.png'))

            # Check if the images are the same
            if ImageChops.difference(image1, image2).getbbox() is not None:

                # Create a new image showing the difference
                diff_image = ImageChops.difference(image1, image2)
                # Save the difference image

                error_folder = Path(test_folder, 'errors')
                pathlib.Path(error_folder).mkdir(parents=True, exist_ok=True)

                diff_image.save(Path(error_folder, f'{name}.png'))
                self.assertTrue(False, f'Page {page_number} is different')

    def test_keep_with_next(self):
        self.run_report(name='keep_with_next')

    def test_basic(self):
        self.run_report(name='basic')

    def test_change_header(self):
        self.run_report(name='change_header')

    def test_border(self):
        self.run_report(name='border')

    def test_background_colour(self):
        self.run_report(name='background_colour')

    def test_abs2(self):
        self.run_report(name='abs2')

    def test_abs3(self):
        self.run_report(name='abs3')

    def test_overflow_gt_height(self):
        self.run_report(name='overflow_gt_height', object_lookup=self.get_sample_objects())

    def test_hidden(self):
        self.run_report(name='hidden')

    @staticmethod
    def get_sample_objects():
        # Define the data for the table
        data = [['Name', 'Age'],
                ['John', 25],
                ['Mary', 30],
                ['Bob', 20],
                ]

        # Define the table style
        table_style = TableStyle([
            ('BACKGROUND', (0, 0), (-1, 0), colors.gray),
            ('TEXTCOLOR', (0, 0), (-1, 0), colors.whitesmoke),
            ('ALIGN', (0, 0), (-1, 0), 'CENTER'),
            ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
            ('FONTSIZE', (0, 0), (-1, 0), 14),
            ('BOTTOMPADDING', (0, 0), (-1, 0), 12),
            ('BACKGROUND', (0, 1), (-1, -1), colors.beige),
            ('TEXTCOLOR', (0, 1), (-1, -1), colors.black),
            ('ALIGN', (0, 1), (-1, -1), 'CENTER'),
            ('FONTNAME', (0, 1), (-1, -1), 'Helvetica'),
            ('FONTSIZE', (0, 1), (-1, -1), 12),
            ('BOTTOMPADDING', (0, 1), (-1, -1), 6),
            ('TOPPADDING', (0, 1), (-1, -1), 6),
            ('GRID', (0, 0), (-1, -1), 1, colors.black),
            ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
            ('ROWBACKGROUNDS', (0, 1), (-1, -1), [colors.whitesmoke, colors.white])
        ])

        # Create the table object
        table = Table(data)

        # Apply the table style
        table.setStyle(table_style)

        return {'sample': table}


import copy
import re
from io import StringIO, BytesIO
from time import sleep

from lxml import etree
from reportlab.lib.colors import HexColor, black
from reportlab.lib.units import mm
from reportlab.platypus import TableStyle, PageBreak, Spacer, Table

from .enhanced_paragraph.enhanced_paragraph import EnhancedParagraph
from .enhanced_paragraph.style import EnhancedParagraphStyle
from .enhanced_table.data import EnhancedTableData
from .enhanced_table.enhanced_tables import OVERFLOW_ROW, EnhancedTable, HEADER_FOOTER, KEEP_TYPE_END, KEEP_TYPE_START, \
    KEEP_TYPE_MIDDLE, KEEP_TYPE_SPAN, KEEP_TYPE_NA
from .png_images import insert_image, insert_obj
from .svglib.svglib import SvgRenderer
from .utils import DocTemplate, get_page_size_from_string, intcomma_currency, ColumnWidthPercentage
from ..pagers.base import BasePager
from ..pagers.border import BorderPager


class ReportXML(object):
    pager_types = {'borders': BorderPager}
    default_pager = BasePager

    entities = [
        (u'lsquo', u'‘'),
        (u'rsquo', u'’'),
        (u'ldquo', u'“'),
        (u'rdquo', u'”'),
        (u'quot', u'"'),
        (u'frasl', u'/'),
        (u'hellip', u'…'),
        (u'ndash', u'–'),
        (u'mdash', u'—'),
        (u'nbsp', u' '),
        (u'not', u'¬'),
        (u'iexcl', u'¡'),
        (u'cent', u'¢'),
        (u'pound', u'£'),
        (u'euro', u'€'),
        (u'curren', u'¤'),
        (u'yen', u'¥'),
        (u'brvbar', u'¦'),
        (u'sect', u'§'),
        (u'uml', u'¨'),
        (u'die', u'¨'),
        (u'copy', u'©'),
        (u'ordf', u'ª'),
        (u'laquo', u'«'),
        (u'reg', u'®'),
        (u'plusmn', u'±'),
        (u'sup2', u'²'),
        (u'sup3', u'³'),
        (u'frac14', u'¼'),
        (u'frac12', u'½'),
        (u'frac34', u'¾'),
        (u'rsquo', u'’'),
    ]

    def __init__(self, object_lookup=None, pager_kwargs=None, test_mode=False, status_method=None):
        self.styles = {}
        if object_lookup is not None:
            self.object_lookup = object_lookup
        else:
            self.object_lookup = {}
        self.background_image_first = None
        self.background_image_remaining = None
        self.background_image_footer = None

        self.styles = {}
        self.border_left_first = 0
        self.border_right_first = 0
        self.border_top_first = 0
        self.border_bottom_first = 0
        self.border_left_continuation = 0
        self.border_right_continuation = 0
        self.border_top_continuation = 0
        self.border_bottom_continuation = 0
        self.page_style = None
        self.test_mode = test_mode
        self.status_method = status_method

        if pager_kwargs is None:
            self.pager_kwargs = {}
        else:
            self.pager_kwargs = pager_kwargs

    def load_xml_and_make_pdf(self, xml, add_doctype=True, background_image_first=None,
                              background_image_remaining=None, background_image_footer=None):

        self.background_image_first = background_image_first
        self.background_image_remaining = background_image_remaining
        self.background_image_footer = background_image_footer

        self.update_status("Loading")
        parser = etree.XMLParser(remove_blank_text=True, resolve_entities=False)

        if add_doctype:
            xml = self.get_doc_type() + xml
        tree = etree.parse(StringIO(xml), parser)
        root = tree.getroot()
        result = BytesIO()
        self.make_pdf(root, result)
        result.seek(0)
        return result

    def get_object(self, object_id):
        if object_id in self.object_lookup.keys():
            return self.object_lookup[object_id]
        else:
            try:
                object_id = int(object_id)
                if object_id in self.object_lookup.keys():
                    return self.object_lookup[object_id]
            except ValueError:
                pass
                # if 'default' in self.object_lookup.keys():
                #     return self.object_lookup['default']
        return None

    def make_pdf(self, root_element, file_buffer):
        title = root_element.get('title')
        page_size = get_page_size_from_string(root_element.get('page_size', ''),
                                              root_element.get('page_orientation', ''))
        pager = self.setup_pager(page_size=page_size,
                                 root_element=root_element)
        doc = DocTemplate(title,
                          pager,
                          file_buffer,
                          pagesize=page_size)
        story = []
        page_width = (pager.page_width() - pager.pageused.left - pager.pageused.right) / mm
        page_height = (pager.page_height() - pager.pageused.top - pager.page_used_bottom()) / mm
        self.process_xml(root_element=root_element,
                         story=story,
                         page_width=page_width,
                         page_height=page_height,
                         top_border=pager.pageused.top,
                         bottom_border=pager.pageused.bottom)
        doc.build(story, canvasmaker=self.canvasmaker)

    def get_pager(self, *args, **kwargs):
        return self.pager_types.get(self.page_style, self.default_pager)(*args, **kwargs)

    def setup_pager(self, page_size, root_element):
        title = root_element.get('title')
        self.page_style = root_element.get('page_style')

        self.border_left_first = float(root_element.get('border_left_first', 0))
        self.border_right_first = float(root_element.get('border_right_first', 0))
        self.border_top_first = float(root_element.get('border_top_first', 0))
        self.border_bottom_first = float(root_element.get('border_bottom_first', 0))

        self.border_left_continuation = float(root_element.get('border_left_continuation',
                                                               self.border_left_first))
        self.border_right_continuation = float(root_element.get('border_right_continuation',
                                                                self.border_right_first))
        self.border_top_continuation = float(root_element.get('border_top_continuation',
                                                              self.border_top_first))
        self.border_bottom_continuation = float(root_element.get('border_bottom_continuation',
                                                                 self.border_bottom_first))

        for name, value in root_element.attrib.items():
            if name.startswith('pager_'):
                name = name[6:]
                if name not in self.pager_kwargs:
                    self.pager_kwargs[name] = value

        return self.get_pager(heading=title,
                              filename=title,
                              pagesize=page_size,
                              border_left_first=self.border_left_first * mm,
                              border_right_first=self.border_right_first * mm,
                              border_top_first=self.border_top_first * mm,
                              border_bottom_first=self.border_bottom_first * mm,
                              border_left_continuation=self.border_left_continuation * mm,
                              border_right_continuation=self.border_right_continuation * mm,
                              border_top_continuation=self.border_top_continuation * mm,
                              border_bottom_continuation=self.border_bottom_continuation * mm,
                              background_image_first=self.background_image_first,
                              background_image_remaining=self.background_image_remaining,
                              background_image_footer=self.background_image_footer,
                              test_mode=self.test_mode,
                              status_method=self.update_status,
                              **self.pager_kwargs)

    def canvasmaker(self, *args, **kwargs):
        return self.get_pager(*args,
                              footer_field='',
                              border_left_first=self.border_left_first * mm,
                              border_right_first=self.border_right_first * mm,
                              border_top_first=self.border_top_first * mm,
                              border_bottom_first=self.border_bottom_first * mm,
                              border_left_continuation=self.border_left_continuation * mm,
                              border_right_continuation=self.border_right_continuation * mm,
                              border_top_continuation=self.border_top_continuation * mm,
                              border_bottom_continuation=self.border_bottom_continuation * mm,
                              background_image_first=self.background_image_first,
                              background_image_remaining=self.background_image_remaining,
                              background_image_footer=self.background_image_footer,
                              test_mode=self.test_mode,
                              status_method=self.update_status,
                              **self.pager_kwargs,
                              **kwargs)

    def process_xml(self, root_element, story, page_width, page_height, top_border, bottom_border):
        self.styles = {}
        children = root_element.getchildren()

        for child in children:
            if child.tag.lower() == "table":
                table = self.process_table(child,
                                           table_width=page_width,
                                           page_height=page_height,
                                           page_width=page_width,
                                           top_border=top_border,
                                           bottom_border=bottom_border)
                if table is not None:
                    story.append(table)

            elif child.tag.lower() == "style":
                self.process_style_element(child.text)
            elif child.tag.lower() == "p":
                story.append(self.process_paragraph_element(child))
            elif child.tag.lower() == "page_break":
                story.append(PageBreak())
            elif child.tag.lower() == "spacer":
                story.append(self.process_spacer_tag(child))

    def process_style_element(self, style_text):
        """
        This implement the style element just like CSS in html does
        :param style_text:
        """
        for match in re.finditer(r"(\w+)[\s\r\n]*{([:;\w\s\r\n-.,]+)\}", style_text):
            style_name = match.group(1)
            style_css = match.group(2)
            style_css = style_css.strip(" \r\n")
            self.styles[style_name] = style_css

    def process_table(self, table, table_width, page_height=None, page_width=None, top_border=None, bottom_border=None):
        """
        This implements tables using EnhanceTable
        :param page_height:
        :param page_width:
        :param table:
        :param table_width:
        :param top_border:
        :param bottom_border:
        """
        main_data = []
        main_styles = []
        main_span = {}

        headers = []
        footers = []

        headers_index = []
        footers_index = []
        current_header_index = None
        current_footer_index = None

        row_heights = []
        keep_data = []
        other_styles = {}
        held_cells = {}

        self.process_css_for_table(table, main_styles, other_styles)

        if self.get_boolean_value(table.get('hidden')):
            return None

        min_rows_top = int(table.get('min_rows_top', 0))
        min_rows_bottom = int(table.get('min_rows_bottom', 0))

        rows_variables = []
        variables = {}

        row_count = -1
        col_widths = []

        layout_widths = table.get('layout_widths')
        if layout_widths is not None:
            for col_width in layout_widths.split(','):
                col_widths.append(self.set_column_width(col_width))

        pos_x = table.get('pos_x')
        pos_y = table.get('pos_y')

        held_row_span = 1
        hidden_columns = set()

        for element in table:
            if element.tag == 'tr':
                row_count += 1
                max_row_span, overflow_row_count = self.process_tr(tr_element=element,
                                                                   data=main_data,
                                                                   styles=main_styles,
                                                                   other_table_styles=other_styles,
                                                                   row_heights=row_heights,
                                                                   row_count=row_count,
                                                                   span=main_span,
                                                                   rows_variables=rows_variables,
                                                                   variables=variables,
                                                                   col_widths=col_widths,
                                                                   table_width=table_width,
                                                                   hidden_columns=hidden_columns,
                                                                   held_cells=held_cells)

                if max_row_span > held_row_span:
                    held_row_span = max_row_span

                if held_row_span > 1 or min_rows_top > 0:
                    min_rows_top -= 1
                    keep_type = KEEP_TYPE_SPAN
                else:
                    keep_type = KEEP_TYPE_NA
                for _ in range(overflow_row_count + 1):
                    keep_data.append(keep_type)
                    headers_index.append(current_header_index)
                    footers_index.append(current_footer_index)

                held_row_span -= 1
                row_count += overflow_row_count
            elif element.tag == 'keep':
                local_keep_data = []
                keep_type = KEEP_TYPE_START
                for child_element in element:
                    if child_element.tag == 'tr':
                        row_count += 1
                        max_row_span, overflow_row_count = self.process_tr(tr_element=child_element,
                                                                           data=main_data,
                                                                           styles=main_styles,
                                                                           other_table_styles=other_styles,
                                                                           row_heights=row_heights,
                                                                           row_count=row_count,
                                                                           span=main_span,
                                                                           rows_variables=rows_variables,
                                                                           variables=variables,
                                                                           col_widths=col_widths,
                                                                           table_width=table_width,
                                                                           hidden_columns=hidden_columns,
                                                                           held_cells=held_cells)
                        if max_row_span > held_row_span:
                            held_row_span = max_row_span
                        if held_row_span > 1 or min_rows_top > 0:
                            min_rows_top -= 1

                        for index in range(overflow_row_count + 1):
                            headers_index.append(current_header_index)
                            footers_index.append(current_footer_index)
                            local_keep_data.append(keep_type)
                            keep_type = KEEP_TYPE_MIDDLE

                        held_row_span -= 1
                        row_count += overflow_row_count

                if len(local_keep_data) > 0:
                    local_keep_data[-1] = KEEP_TYPE_END
                keep_data += local_keep_data

            elif element.tag == 'no_headers':
                current_header_index = None

            elif element.tag == 'no_footers':
                current_footer_index = None

            elif element.tag in ['header', 'footer']:
                is_header = element.tag == 'header'
                output = element.get('output', "0")
                if is_header and self.get_boolean_value(output):
                    for tr in element:
                        row_count += 1
                        _, overflow_row_count = self.process_tr(tr_element=tr,
                                                                data=main_data,
                                                                styles=main_styles,
                                                                other_table_styles=other_styles,
                                                                row_heights=row_heights,
                                                                row_count=row_count,
                                                                span=main_span,
                                                                rows_variables=rows_variables,
                                                                variables=variables,
                                                                held_cells=held_cells,
                                                                col_widths=col_widths,
                                                                hidden_columns=hidden_columns,
                                                                table_width=table_width)
                header_footer_span = {}
                header_footer_data = []
                header_footer_commands = []
                header_footer_row_height = []
                temp_rows_variables = []

                for row_index, tr in enumerate(element, HEADER_FOOTER):
                    self.process_tr(tr_element=tr,
                                    data=header_footer_data,
                                    styles=header_footer_commands,
                                    other_table_styles=other_styles,
                                    row_heights=header_footer_row_height,
                                    row_count=row_index,
                                    span=header_footer_span,
                                    rows_variables=temp_rows_variables,
                                    variables=variables,
                                    col_widths=col_widths,
                                    table_width=table_width,
                                    default_row_height=35,
                                    is_header_or_footer=True,
                                    hidden_columns=hidden_columns)

                enhanced_table_data = EnhancedTableData(row_data=header_footer_data,
                                                        row_heights=header_footer_row_height,
                                                        commands=header_footer_commands)

                if is_header:
                    current_header_index = len(headers)
                    headers.append(enhanced_table_data)
                else:
                    current_footer_index = len(footers)
                    footers.append(enhanced_table_data)

        length = len(keep_data)
        for x in range(1, min_rows_bottom + 1):
            if length - x < 0:
                break
            keep_data[length - x] = KEEP_TYPE_SPAN

        h_align, v_align = self.get_alignment_details(main_styles)

        new_column_widths = self.process_column_widths(col_widths, table_width)

        if main_data:
            t = EnhancedTable(table_data={'row_data': main_data,
                                          'row_variables': rows_variables,
                                          'keep_with_next': keep_data,
                                          'headers_index': headers_index,
                                          'footers_index': footers_index},
                              row_heights=row_heights,
                              repeat_rows=0,
                              headers=headers,
                              footers=footers,
                              h_align=h_align,
                              v_align=v_align,
                              col_widths=new_column_widths,
                              initial=True)
        else:
            return None

        if pos_y is not None and pos_x is not None and page_height is not None:
            ref_is_top = table.get('pos_y_ref', 'top') == 'top'
            ref_is_right = table.get('pos_x_ref', 'left') == 'right'
            ignore_margin = table.get('ignore_margin', 'no') == 'yes'
            if ref_is_top:
                page_height *= mm
                page_height += top_border + bottom_border
                height = t.get_height()
                new_pos_y = page_height - (float(pos_y) + height)
                if not ignore_margin and top_border is not None:
                    new_pos_y -= float(top_border)
            else:
                new_pos_y = float(pos_y)
                if not ignore_margin and bottom_border is not None:
                    new_pos_y += float(bottom_border)
            t.pos_y = str(new_pos_y)

            if ref_is_right:
                page_width *= mm
                width = t.get_width()
                new_pos_x = page_width - (float(pos_x) + width)
            else:
                new_pos_x = pos_x
            t.pos_x = str(new_pos_x)

        t.setStyle(TableStyle(main_styles))

        return t

    @staticmethod
    def coord(x, y):
        """
        http://stackoverflow.com/questions/4726011/wrap-text-in-a-table-reportlab
        Helper class to help position flowables in Canvas objects
        """
        x, y = x * mm, 222 - y * mm
        return x, y

    def process_tr(self, tr_element, data, styles, other_table_styles,
                   row_heights, row_count, span, rows_variables, variables, col_widths, table_width, hidden_columns,
                   held_cells=None, default_row_height=None, is_header_or_footer=False):

        max_row_span = 0
        if self.get_boolean_value(tr_element.get('hidden')):
            return max_row_span, 0

        row_data = []
        other_styles = {}

        self.process_css_for_table(tr_element, styles, other_styles, start_row=row_count, end_row=row_count)
        offset = 0
        max_row_span = 0

        overflow_rows = []
        overflow_row_count = 0

        row_variables = copy.copy(variables)
        for variables_element in tr_element.iter('variables'):
            for variable in variables_element.attrib:
                value = variables_element.get(variable)
                variables[variable] = value
                row_variables[variable] = value

        for variable_element in tr_element.iter('variable'):
            variable = variable_element.attrib.get('name', '')
            value = variable_element.text
            variables[variable] = value
            row_variables[variable] = value

        for variables_element in tr_element.iter('currency_variables'):
            symbol = variables_element.attrib.get('symbol', '')

            for variable in variables_element.attrib:
                if variable == 'symbol':
                    continue
                value = float(variables_element.get(variable))
                variables[variable] = value
                row_variables[variable] = value

                symbol_name = variable + '__symbol'
                variables[symbol_name] = symbol
                row_variables[symbol_name] = symbol

                currency_variable = variable + '__currency'
                currency_value = symbol + intcomma_currency(value)

                variables[currency_variable] = currency_value
                row_variables[currency_variable] = currency_value

        for variables_element in tr_element.iter('variable_addition'):
            for variable in variables_element.attrib:
                try:
                    if variables.get(variable) is None:
                        variables[variable] = float(variables_element.get(variable))
                    else:
                        variables[variable] = float(variables.get(variable)) + float(variables_element.get(variable))
                except ValueError:
                    pass
                row_variables[variable] = variables.get(variable)
                currency_variable = variable + '__currency'
                if currency_variable in variables:
                    symbol_name = variables[variable + '__symbol']

                    currency_value = symbol_name + intcomma_currency(float(variables.get(variable)) / 100.0)

                    row_variables[currency_variable] = currency_value
                    variables[currency_variable] = currency_value

        rows_variables.append(row_variables)
        col_count = 0
        for index, td_element in enumerate(tr_element, 0):

            if td_element.tag != 'td':
                continue

            name = td_element.get('name')
            if name is not None and name in hidden_columns:
                continue
            if index in hidden_columns:
                continue

            hidden_column = td_element.get('hidden_column')
            col_span = int(td_element.get('colspan', "1"))
            row_span = int(td_element.get('rowspan', "1"))

            if self.get_boolean_value(hidden_column):
                if name is not None:
                    hidden_columns.add(name)
                else:
                    for h_index in range(col_span):
                        hidden_columns.add(index + h_index)
                continue
            if td_element.get('hidden'):
                continue

            if len(col_widths) < col_count + 1:
                col_widths.append(None)

            # check that cell is not marked as a rowspan
            p_offset = span.get('%d-%d' % (row_count, col_count + offset), None)
            while p_offset is not None and p_offset > 0:
                for x in range(0, p_offset):
                    row_data.append('')
                offset += p_offset
                p_offset = span.get('%d-%d' % (row_count, col_count + offset), None)

            if row_span > max_row_span:
                max_row_span = row_span

            # get the styles for td
            self.process_css_for_table(td_element, styles, other_styles,
                                       start_col=col_count + offset, start_row=row_count,
                                       end_col=col_count + offset + col_span - 1, end_row=row_count + row_span - 1)

            if len(td_element) > 0 and td_element[0].tag == 'table':
                new_table_column_widths = self.process_column_widths(col_widths, table_width)
                new_table_column_width = 0
                new_table_column_widths_count = len(new_table_column_widths)
                if index < new_table_column_widths_count:
                    for col_num in range(index, index + col_span):
                        if col_num > new_table_column_widths_count:
                            break
                        try:
                            new_table_column_width += new_table_column_widths[col_num + offset] / mm
                        except IndexError:
                            pass

                else:
                    new_table_column_width = new_table_column_widths[0] / mm
                padding = int(self.get_padding_for_cell(styles,
                                                        start_col=col_count + offset,
                                                        start_row=row_count,
                                                        end_col=-col_count + offset + col_span - 1,
                                                        end_row=-row_count + row_span - 1) / mm)
                display_object = self.process_table(td_element[0], new_table_column_width - padding)
                if display_object is None:
                    display_object = ''
            elif len(td_element) > 0 and td_element[0].tag[-3:] == 'svg':
                display_object = self.svg2rlg_from_node(td_element[0])
            elif len(td_element) > 0 and td_element[0].tag[-3:] == 'png':
                display_object = insert_image(td_element[0])
            elif len(td_element) > 0 and td_element[0].tag[-3:] == 'obj':
                object_id = td_element[0].get('id', "")
                if object_id != "":
                    display_object = self.get_object(object_id)
                else:
                    display_object = insert_obj(td_element[0])
            elif len(td_element) > 0 and td_element[0].tag[-12:] == 'currency_qty':
                variable_name = td_element[0].get('variable')
                qty = int(td_element[0].get('qty'))

                if variable_name is not None:
                    symbol = ''
                    value = variables[variable_name]
                    if symbol == '':
                        symbol = variables.get(variable_name + '__symbol', '')
                else:
                    symbol = td_element[0].get('symbol', '')
                    if symbol == '':
                        symbol_from = td_element[0].get('symbol_from', '')
                        symbol = variables.get(symbol_from + '__symbol', '')
                    value = td_element[0].get('value')

                unit = (float(value) / qty) / 100.0

                display_str = '%s%s' % (symbol, intcomma_currency(unit, show_more_decimal_places=True))

                style = self.process_css_for_table_paragraph_style(styles, row_count, col_count + offset)
                display_object = EnhancedParagraph(display_str, style, css_classes=self.styles)

            elif len(td_element) > 0 and td_element[0].tag[-8:] == 'currency':
                variable_name = td_element[0].get('variable')
                if is_header_or_footer:
                    variable_name = td_element[0].get('variable')
                    if variable_name is not None:
                        display_object = '%%(%s__currency)s' % variable_name
                    else:
                        value = int(td_element[0].get('value'))
                        number_string = "%.2f" % (float(value) / 100.0)
                        display_object = '%s' % (intcomma_currency(number_string))

                else:
                    if variable_name is not None:
                        symbol = ''
                        value = variables.get(variable_name)
                        if symbol == '':
                            symbol = variables.get(variable_name + '__symbol', '')
                    else:
                        symbol = td_element[0].get('symbol', '')
                        value = td_element[0].get('value')

                    add_to_variable_name = td_element[0].get('add_to')

                    if add_to_variable_name is not None:
                        if symbol == '':
                            symbol = variables.get(add_to_variable_name + '__symbol', '')

                        variables[add_to_variable_name] += float(value)
                        row_variables[add_to_variable_name] = variables.get(add_to_variable_name)
                        currency_variable = add_to_variable_name + '__currency'
                        if currency_variable in variables:
                            currency_value = symbol + intcomma_currency(
                                float(variables.get(add_to_variable_name)) / 100.0)

                            row_variables[currency_variable] = currency_value
                            variables[currency_variable] = currency_value
                    if value is not None:
                        number_string = "%.2f" % (float(value) / 100.0)
                        display_object = '%s%s' % (symbol, intcomma_currency(number_string))
                    else:
                        display_object = ''

            elif len(td_element) > 0:
                xml = etree.tostring(td_element, pretty_print=True)

                overflow_gt_height = td_element.get('overflow_gt_height')
                overflow_gt_length = int(td_element.get('overflow_gt_length', 0))
                style = self.process_css_for_table_paragraph_style(styles, row_count, col_count + offset)

                if overflow_gt_height is not None:

                    display_object, overflow_row_count = self.process_overflow_height(
                        td_element=td_element,
                        index=index,
                        overflow_gt_height=overflow_gt_height,
                        xml=xml,
                        style=style,
                        col_widths=col_widths,
                        table_width=table_width,
                        col_span=col_span,
                        styles=styles,
                        offset=offset,
                        col_count=col_count,
                        row_count=row_count,
                        row_span=row_span,
                        row_data=row_data,
                        overflow_rows=overflow_rows,
                        rows_variables=rows_variables,
                        held_cells=held_cells)

                elif overflow_gt_length:
                    out_xml, style, overflow_row_count = self.overflow_cell(td_element=td_element,
                                                                            xml=xml,
                                                                            overflow_gt_length=overflow_gt_length,
                                                                            styles=styles,
                                                                            style=style,
                                                                            offset=offset,
                                                                            col_count=col_count,
                                                                            row_count=row_count,
                                                                            col_span=col_span,
                                                                            row_span=row_span,
                                                                            row_data=row_data,
                                                                            overflow_rows=overflow_rows,
                                                                            rows_variables=rows_variables)
                    display_object = EnhancedParagraph(out_xml, style, css_classes=self.styles)
                else:
                    display_object = EnhancedParagraph(xml, style, css_classes=self.styles)

            else:
                display_object = td_element.text
                if display_object is None:
                    display_object = ''

            hold_cell = td_element.get('hold_cell')
            if hold_cell is not None:
                held_cells[hold_cell] = {'display_object': display_object, 'index': index, 'colspan': col_span}
            row_data.append(display_object)

            width = self.set_column_width(td_element.get('width'))
            if width is not None:
                col_widths[col_count] = width

            if col_span > 1 or row_span > 1:
                for x in range(1, col_span):
                    row_data.append('')
                styles.append(('SPAN',
                               (col_count + offset, row_count),
                               (col_count + offset + col_span - 1, row_count + row_span - 1)))
                for x in range(1, overflow_row_count+1):
                    styles.append(('SPAN',
                                   (col_count + offset, x + row_count),
                                   (col_count + offset + col_span - 1, x + row_count + row_span - 1)))

                if row_span > 1:
                    for r in range(1, row_span):
                        span['%d-%d' % (r + row_count + overflow_row_count, col_count + offset)] = col_span
                offset += col_span - 1
            col_count += 1

        data.append(row_data)

        row_height = self.get_row_height(other_table_styles, other_styles)
        if row_height is None:
            row_height = tr_element.get('row_height', default_row_height)
            if row_height is not None:
                row_height = float(row_height)

        row_heights.append(row_height)

        for overflow_row in overflow_rows:
            data.append(overflow_row)
            row_heights.append(OVERFLOW_ROW)

        return max_row_span, overflow_row_count

    @staticmethod
    def set_column_width(col_width):
        if col_width is not None and col_width != '':
            if col_width[-1] == '%':
                width = ColumnWidthPercentage(float(col_width[:-1]))
            else:
                width = float(col_width)
            return width
        return None

    @staticmethod
    def get_row_height(other_table_styles, other_styles):
        row_height = other_styles.get('ROW_HEIGHT')
        if row_height is None:
            row_height = other_table_styles.get('ROW_HEIGHT')

        if row_height is not None:
            row_height = row_height * mm

        return row_height

    def process_css_for_table(self, tag, styles, other_styles, start_col=0, start_row=0, end_col=-1, end_row=-1,
                              style_tag_name='style', class_tag_name='class'):

        css = self.get_css_from_style_attribute(tag=tag,
                                                style_tag_name=style_tag_name,
                                                class_tag_name=class_tag_name)
        if css == '':
            return

        self.convert_css_to_style(css, styles, other_styles, start_col, start_row, end_col, end_row)

    @staticmethod
    def convert_css_to_style(css, styles, other_styles, start_col=0, start_row=0, end_col=-1, end_row=-1):
        start_tuple = (start_col, start_row)
        end_tuple = (end_col, end_row)
        styles_list = css.split(';')
        for style in styles_list:
            if style == '':
                continue
            style_type, style_detail = style.split(':')
            style_type = style_type.lower().lstrip("\r\n ")

            if style_type in ('inner_grid',
                              'box',
                              'line_above',
                              'line_below',
                              'line_before',
                              'line_after'):
                details = style_detail.split(',')
                if len(details) > 3:

                    styles.append((style_type.replace("_", "").upper(),
                                   (0, int(float(details[2]))), (-1, float(details[3]),
                                                                 float(details[0]), HexColor(details[1]))))
                elif len(details) > 2:
                    styles.append((style_type.replace("_", "").upper(), (0, int(float(details[2]))), end_tuple,
                                   float(details[0]), HexColor(details[1])))

                elif len(details) > 1:
                    styles.append((style_type.replace("_", "").upper(), start_tuple, end_tuple,
                                   float(details[0]), HexColor(details[1])))
                else:
                    styles.append((style_type.replace("_", "").upper(), start_tuple, end_tuple,
                                   float(details[0]), black))
            elif style_type in ('text_color',
                                'background'):
                styles.append((style_type.replace("_", "").upper(), start_tuple, end_tuple, HexColor(style_detail)))
            elif style_type in ('valign',
                                'halign',
                                'align'):
                styles.append((style_type.replace("_", "").upper(), start_tuple, end_tuple, style_detail.upper()))
            elif style_type in ('font',
                                'face',
                                'font_name'):
                styles.append((style_type.replace("_", "").upper(), start_tuple, end_tuple, style_detail))

            elif style_type in ('left_padding',
                                'right_padding',
                                'bottom_padding',
                                'top_padding'):
                styles.append((style_type.replace("_", "").upper(), start_tuple, end_tuple, float(style_detail) * mm))

            elif style_type in ('leading',
                                'font_size',
                                'size'):
                styles.append((style_type.replace("_", "").upper(), start_tuple, end_tuple, int(style_detail)))
            elif style_type == 'row_height':
                other_styles[style_type.upper()] = int(style_detail)

    @staticmethod
    def get_padding_for_cell(styles, start_col=0, start_row=0, end_col=-1, end_row=-1):
        start_tuple = (start_col, start_row)
        padding = 0
        found_padding = False
        for style in styles:

            style_type = style[0]
            if style_type in ('LEFTPADDING',
                              'RIGHTPADDING'):

                if start_tuple == style[1]:  # todo this isn't always going to work if defined on tr or table
                    padding += style[3]
                    found_padding = True
        if found_padding:
            return padding - 1
        return 3

    def get_css_from_style_attribute(self, tag, style_tag_name='style', class_tag_name='class'):

        style_tag = tag.get(style_tag_name)
        class_tag = tag.get(class_tag_name)

        if style_tag is None and class_tag is None:
            return ''
        css = ''
        if class_tag is not None and self.styles.get(class_tag) is not None:
            css = self.styles.get(class_tag)
        if style_tag is not None:
            css += style_tag
        return css

    @staticmethod
    def process_css_for_table_paragraph_style(css, row_count, col_count):
        paragraph_style = EnhancedParagraphStyle('paragraph_style')
        paragraph_style.process_css_for_table_paragraph_style(css, row_count, col_count)
        return paragraph_style

    def process_paragraph_element(self, tag):

        css = self.get_css_from_style_attribute(tag)
        paragraph_style = EnhancedParagraphStyle('paragraph_style')
        paragraph_style.process_raw_css(css)
        xml = etree.tostring(tag)

        enhanced_paragraph = EnhancedParagraph(xml, paragraph_style, css_classes=self.styles)
        return enhanced_paragraph

    @staticmethod
    def get_alignment_details(main_styles):
        h_align = 'LEFT'
        v_align = 'TOP'

        for style in main_styles:
            if style[0].lower() == 'halign':
                h_align = style[3]
            if style[0].lower() == 'valign':
                v_align = style[3]
        return h_align, v_align

    @staticmethod
    def process_column_widths(col_widths, table_width):
        new_col_widths = copy.copy(col_widths)
        undefined_count = 0
        defined_percentage = 0
        defined_space = 0

        for index, x in enumerate(new_col_widths, 0):
            if x is None:
                undefined_count += 1
            elif isinstance(x, ColumnWidthPercentage):
                defined_percentage += x.get_value()
            else:
                defined_space += x
                new_col_widths[index] = x * mm

        if undefined_count == 0 and defined_percentage == 0:
            return new_col_widths

        undefined_percentage = 0
        if undefined_count > 0:
            undefined_percentage = (100 - defined_percentage) / undefined_count

        available_space = table_width - defined_space
        if available_space < 0:
            available_space = 0

        percentage_amount = available_space / 100.0

        for index, x in enumerate(new_col_widths, 0):
            if x is None:
                value = undefined_percentage * percentage_amount
                new_col_widths[index] = value * mm
            elif isinstance(x, ColumnWidthPercentage):
                new_col_widths[index] = (x.get_value() * percentage_amount) * mm

        return new_col_widths

    def process_spacer_tag(self, element):

        css = self.get_css_from_style_attribute(element)
        css = css.replace('\r', '').replace('\n', '')
        height = 10
        styles_list = css.split(';')
        for style in styles_list:
            if style == '':
                continue
            style_type, style_detail = style.split(':')
            style_type = style_type.lower().lstrip("\r\n ")
            if style_type == 'height':
                height = float(style_detail)

        return Spacer(0, height * mm)

    @staticmethod
    def svg2rlg_from_node(node):
        svg_renderer = SvgRenderer(path=None)
        drawing = svg_renderer.render(node)
        return drawing

    @staticmethod
    def split_cell(xml, overflow_length):
        overflow_rows = []
        held_working_tags = []
        raw_parts = None
        offset = xml.find('>'.encode())
        m = re.search("<br\s?/>".encode(), xml[overflow_length + offset:])
        if m:
            pos = m.start() + overflow_length + offset
            span = m.span()
            match_len = span[1] - span[0]
            xml_parts = re.split("<br\s?/>".encode(), xml[pos + match_len:])
            next_xml = ''
            raw_parts = []
            # this makes the xml valid again after it has been split

            for xml_part in [xml[:pos]] + xml_parts:
                working_xml = next_xml.encode() + xml_part
                tags = re.findall(r'<[^<>]*>', working_xml.decode())
                next_xml = ''
                working_tags = []
                for tag in tags:
                    if tag[-2] == '/':
                        continue
                    m = re.search(r'\w+', tag)
                    tag_name = m.group(0)

                    if tag[1] == '/':
                        if len(working_tags) == 0:
                            break
                        elif tag_name == working_tags[-1][0]:
                            working_tags.pop()
                            if tag_name == 'td' and len(working_tags) == 0:
                                break
                    else:
                        working_tags.append((tag_name, tag))
                raw_parts.append(xml_part)

                held_working_tags.append(working_tags)
                for tag in reversed(working_tags):
                    working_xml += ('</' + tag[0] + '>').encode()
                overflow_rows.append(working_xml)
                for tag in working_tags:
                    next_xml += tag[1]

            return overflow_rows[0], overflow_rows[1:], raw_parts, held_working_tags

        return xml, overflow_rows, raw_parts, held_working_tags

    def overflow_cell(self, td_element, xml, overflow_gt_length, styles, style, offset, col_count, row_count, col_span,
                      row_span, row_data, overflow_rows, rows_variables):
        text = re.sub('<.*?>', '', str(xml))
        if len(text) > overflow_gt_length:
            overflow_length = int(td_element.get('overflow_length', 0))
            xml, overflow_rows_xml, _, _ = self.split_cell(xml, overflow_length)

            if overflow_rows_xml:
                self.process_css_for_table(tag=td_element,
                                           styles=styles,
                                           other_styles={},
                                           start_col=col_count + offset,
                                           start_row=row_count,
                                           end_col=col_count + offset + col_span - 1,
                                           end_row=row_count + row_span - 1,
                                           style_tag_name='overflow_top_style',
                                           class_tag_name='overflow_top_class')

                self.process_over_flow_xml(overflow_rows_xml=overflow_rows_xml,
                                           overflow_rows=overflow_rows,
                                           rows_variables=rows_variables,
                                           td_element=td_element,
                                           styles=styles,
                                           row_count=row_count,
                                           col_count=col_count,
                                           row_span=row_span,
                                           col_span=col_span,
                                           row_data=row_data,
                                           offset=offset)

                style = self.process_css_for_table_paragraph_style(styles, row_count, col_count + offset)
        return xml, style, len(overflow_rows)

    def get_doc_type(self):
        entity_string = ''
        for entity in self.entities:
            entity_string += u'<!ENTITY %s \'%s\'>' % (entity[0], entity[1])
        xml = """<!DOCTYPE root SYSTEM "print_engine" [%s]>""" % entity_string
        return xml

    @staticmethod
    def get_overflow_overflow_gt_height(raw_overflow_gt_height, held_cells, widths):
        if raw_overflow_gt_height.startswith('#'):
            held_cell_data = held_cells.get(raw_overflow_gt_height[1:])
            display_object = held_cell_data['display_object']
            index = held_cell_data['index']
            col_span = held_cell_data['colspan']
            avail_width = sum(widths[index:index + col_span])

            if isinstance(display_object, EnhancedParagraph):
                return display_object.calc_text_height(avail_width=avail_width)
            elif isinstance(display_object, Table):
                _, h = display_object.wrap(0, 0)
                return h
            else:
                return 5 * mm
        else:
            return float(raw_overflow_gt_height) * mm

    def process_overflow_height(self, td_element, index, overflow_gt_height, xml, style,
                                col_widths, table_width, col_span,
                                styles, offset, col_count, row_count, row_span, row_data, overflow_rows,
                                rows_variables, held_cells):
        widths = self.process_column_widths(col_widths, table_width)
        overflow_gt_height = self.get_overflow_overflow_gt_height(raw_overflow_gt_height=overflow_gt_height,
                                                                  held_cells=held_cells,
                                                                  widths=widths)

        avail_width = sum(widths[index:index+col_span])
        display_object = EnhancedParagraph(xml, style, css_classes=self.styles)
        height = display_object.calc_text_height(avail_width=avail_width)

        if height <= overflow_gt_height:
            return display_object, 0

        xml, overflow_rows_xml, raw_parts, held_working_tags = self.split_cell(xml=xml, overflow_length=0)

        working_height = 0
        working_xml = b''
        found_break = False
        overflow_row_offset = 0
        for overflow_row_offset, (row_xml, raw_part, working_tags) in\
                enumerate(zip([xml] + overflow_rows_xml, raw_parts, held_working_tags), 1):
            working_xml += raw_part + '<br/>'.encode()
            if raw_part == b'\n':
                row_xml = row_xml.replace(b'\n', b'M')
            p = EnhancedParagraph(row_xml, style, css_classes=self.styles)
            height = p.calc_text_height(avail_width=avail_width)
            working_height += height
            if working_height > overflow_gt_height:
                for tag in reversed(working_tags):
                    working_xml += ('</' + tag[0] + '>').encode()
                    found_break = True
                break
        overflow_len = len(overflow_rows_xml[overflow_row_offset-1:])
        if not found_break or overflow_len == 0:
            return display_object, 0

        display_object = EnhancedParagraph(working_xml, style, css_classes=self.styles)

        self.process_css_for_table(tag=td_element,
                                   styles=styles,
                                   other_styles={},
                                   start_col=col_count + offset,
                                   start_row=row_count,
                                   end_col=col_count + offset + col_span - 1,
                                   end_row=row_count + row_span - 1,
                                   style_tag_name='overflow_top_style',
                                   class_tag_name='overflow_top_class')

        over_flow_len = self.process_over_flow_xml(overflow_rows_xml=overflow_rows_xml[overflow_row_offset - 1:],
                                                   overflow_rows=overflow_rows,
                                                   rows_variables=rows_variables,
                                                   td_element=td_element,
                                                   styles=styles,
                                                   row_count=row_count,
                                                   col_count=col_count,
                                                   row_span=row_span,
                                                   col_span=col_span,
                                                   row_data=row_data,
                                                   offset=offset)
        return display_object, over_flow_len

    def process_over_flow_xml(self, overflow_rows_xml, overflow_rows, rows_variables,
                              td_element, styles, row_count, col_count, row_span, col_span, row_data, offset):

        for overflow_row_offset, row_xml in enumerate(overflow_rows_xml, 1):
            if row_xml == overflow_rows_xml[-1]:
                style_tag_name = 'bottom'
            else:
                style_tag_name = 'middle'

            self.process_css_for_table(tag=td_element,
                                       styles=styles,
                                       other_styles={},
                                       start_col=col_count + offset,
                                       start_row=row_count + overflow_row_offset,
                                       end_col=col_count + offset + col_span - 1,
                                       end_row=row_count + overflow_row_offset + row_span - 1)
            self.process_css_for_table(tag=td_element,
                                       styles=styles,
                                       other_styles={},
                                       start_col=col_count + offset,
                                       start_row=row_count + overflow_row_offset,
                                       end_col=col_count + offset + col_span - 1,
                                       end_row=row_count + overflow_row_offset + row_span - 1,
                                       style_tag_name='overflow_%s_style' % style_tag_name,
                                       class_tag_name='overflow_%s_class' % style_tag_name)
            style = self.process_css_for_table_paragraph_style(
                css=styles,
                row_count=row_count + overflow_row_offset,
                col_count=col_count + offset)
            overflow_object = EnhancedParagraph(row_xml, style, css_classes=self.styles)
            overflow_row = ['' for _ in range(len(row_data))]
            # noinspection PyTypeChecker
            overflow_row.append(overflow_object)
            overflow_rows.append(overflow_row)
            rows_variables.append(rows_variables[-1])

        return len(overflow_rows_xml)

    # noinspection PyMethodMayBeStatic
    def get_boolean_value(self, value):
        return value in ['1', 'True', 'yes']

    def update_status(self, message):
        if self.status_method is not None:
            self.status_method(message)



from django.db import models
from django.template import Template, Context

from django_advanced_pdf.engine.report_xml import ReportXML


class PrintingTemplate(models.Model):
    name = models.CharField(max_length=128, unique=True)
    xml = models.TextField()

    def __str__(self):
        return self.name

    def make_pdf(self, context=None, add_doctype=True, object_lookup=None,
                 background_image_first=None, background_image_remaining=None, background_image_footer=None, **kwargs):
        if context is None:
            xml = self.xml
        else:
            t = Template(self.xml)
            c = Context(context)
            xml = t.render(c)
        report_xml = ReportXML(object_lookup=object_lookup, pager_kwargs=kwargs)
        return report_xml.load_xml_and_make_pdf(xml,
                                                add_doctype=add_doctype,
                                                background_image_first=background_image_first,
                                                background_image_remaining=background_image_remaining,
                                                background_image_footer=background_image_footer)



from reportlab.platypus import Paragraph


class DataParagraph(Paragraph, object):
    """
    Helper class for use in EnhancedTable footer/header row data to allow styles to be applied after variable
    substitution.
    When encountered in a footer or header, the row's data variables are merged into the text and the whole thing then
    converted into a Paragraph, using the styles supplied.
    An optional step after data merging is to run postprocess_fn on the resulting text. This allows for e.g. limited
    reformatting or stripping of characters
    """

    def __init__(self, text, style, postprocess_fn=None):
        """
        @type text: str
        @param text: The text (optionally a fornmat string) for the cell paragraph
        @param style: the cell style to apply to the enclosing paragraph
        @type postprocess_fn: callable
        @param postprocess_fn: An optional function to call after variable values have been substituted into the text
        """
        self.text = text
        self.style = style
        self.postprocess_fn = postprocess_fn

    def merge_variables(self, variables=None):
        if variables:
            new_text = self.text % variables
        else:
            new_text = self.text
        if callable(self.postprocess_fn):
            return self.postprocess_fn(new_text, self.style)
        else:
            return Paragraph(new_text, self.style)



from reportlab.lib.colors import HexColor
from reportlab.platypus import ParaParser
from reportlab.platypus.paraparser import _lineRepeats


class EnhancedParaParser(ParaParser):

    def __init__(self, css_classes, verbose=0, case_sensitive=0, ignore_unknown_tags=1):
        self.css_classes = css_classes
        ParaParser.__init__(self, verbose, case_sensitive, ignore_unknown_tags)

    def start_span(self, attributes):

        class_name = attributes.get('class')
        css = ""
        if class_name is not None:
            css = self.css_classes.get(class_name, '')

        css += attributes.get("style", '')
        styles_list = css.split(';')
        styles = {}
        for style in styles_list:
            if style == '':
                continue
            style_type, style_detail = style.split(':')
            style_type = style_type.lower().lstrip("\r\n ")
            if style_type in ('font',
                              'face',
                              'font_name'):
                styles['fontName'] = style_detail
            elif style_type in ('font_size',
                                'size'):
                styles['fontSize'] = int(style_detail)
            elif style_type in ('text_color',):
                styles['textColor'] = HexColor(style_detail)
            elif style_type == 'text-decoration' and 'underline' in style_detail:
                frag = self._stack[-1]
                styles['us_lines'] = [(
                    self.nlines,
                    'underline',
                    getattr(frag, 'underlineColor', None),
                    getattr(frag, 'underlineWidth', '1'),
                    getattr(frag, 'underlineOffset', self._defaultLineOffsets['underline']),
                    frag.rise,
                    _lineRepeats[getattr(frag, 'underlineKind', 'single')],
                    getattr(frag, 'underlineGap', self._defaultLineGaps['underline']),
                )]

        self._push('span', **styles)

    def end_span(self):
        self._pop('span')



from reportlab.lib import colors
from reportlab.lib.units import mm
from reportlab.lib.utils import ImageReader
from reportlab.pdfgen import canvas

from django_advanced_pdf.engine.utils import PageUsed


class BasePager(canvas.Canvas):

    def __init__(self, heading, filename, pagesize=None, bottom_up=1,
                 page_compression=None, invariant=None, verbosity=0,
                 encrypt=None, crop_marks=None, pdf_version=None, enforce_color_space=None,
                 border_left_first=0, border_right_first=0,
                 border_top_first=0, border_bottom_first=0,
                 border_left_continuation=0, border_right_continuation=0,
                 border_top_continuation=0, border_bottom_continuation=0,
                 program_name=None,
                 background_image_first=None, background_image_remaining=None,
                 background_image_footer=None, test_mode=False, status_method=None, **kwargs):

        self.status_method = status_method
        self.heading = heading
        self.pagesize = pagesize
        self.drawmethods = []
        self.image2 = None
        self._saved_page_states = []
        self.test_mode = test_mode

        self.pageused = PageUsed(left=border_left_first,
                                 right=border_right_first,
                                 top=border_top_first,
                                 bottom=border_bottom_first)
        self.pageused2 = PageUsed(left=border_left_continuation,
                                  right=border_right_continuation,
                                  top=border_top_continuation,
                                  bottom=border_bottom_continuation)
        self.program_name = program_name

        self.background_image_first = background_image_first
        self.background_image_remaining = background_image_remaining
        self.background_image_footer = background_image_footer

        canvas.Canvas.__init__(self, filename, pagesize, bottom_up, page_compression, invariant, verbosity, encrypt,
                               crop_marks, pdf_version, enforce_color_space)

        self.add_background_images()

    def inkAnnotation(self, contents, ink_list=None, rect=None, add_to_page=1, name=None, relative=0, **kw):
        pass

    def add_background_images(self):
        if self.background_image_first is not None or self.background_image_remaining is not None:
            self.draw_first_page_background()
            self.draw_footer_image_block()

    def add_draw_method(self, method):

        """
        Add a drawing method

        @type   method  : Bound method
        @param  method  : Drawing method
        """

        self.drawmethods.append(method)

    def set_page_used(self, left=0, right=0, top=0, bottom=0):
        if left > self.pageused.left:
            self.pageused.left = left
            self.pageused2.left = left
        if right > self.pageused.right:
            self.pageused.right = right
            self.pageused2.right = right
        if top > self.pageused.top:
            self.pageused.top = top
            self.pageused2.top = top
        if bottom > self.pageused.bottom:
            self.pageused.bottom = bottom
            self.pageused2.bottom = bottom

    def page_used_left(self, first_page=True):
        return self.pageused.left if first_page else self.pageused2.left

    def page_used_right(self, first_page=True):
        return self.pageused.right if first_page else self.pageused2.right

    def page_used_top(self, first_page=True):
        return self.pageused.top if first_page else self.pageused2.top

    def page_used_bottom(self, first_page=True):
        return self.pageused.bottom if first_page else self.pageused2.bottom

    def page_width(self):
        return self.pagesize[0]

    def page_height(self):
        return self.pagesize[1]

    def border_left(self, first_page=True):
        return self.page_used_left(first_page)

    def border_right(self, first_page=True):
        return self.page_width() - self.page_used_right(first_page)

    def border_top(self, first_page=True):
        return self.page_height() - self.page_used_top(first_page)

    def border_bottom(self, first_page=True):
        return self.page_used_bottom(first_page)

    def draw_borders(self, page_count):
        """
        Draw the footer block.
        numbers.

        @type   page_count  : int
        @param  page_count  : Total number of pages
        """
        self.display_program_name()

        self.setFont("Helvetica", 10)

        self.drawRightString(self.border_right(self._pageNumber == 1),
                             self.border_bottom(self._pageNumber == 1) - 4 * mm,
                             "Page %d of %d" %
                             (self._pageNumber, page_count))

    def display_program_name(self):
        if self.program_name is not None:
            border_left = 15
            border_bottom = 4
            self.setFont("Helvetica", 4)

            self.drawString(border_left,
                            border_bottom,
                            self.program_name)

    def draw_page_number(self, page_count):

        """
        @type   page_count  : int
        @param  page_count  : Total number of pages
        """

        for method in self.drawmethods:
            method(page_count)
        self.draw_borders(page_count)

        #  Enable the condition below to draw a red box round the margins.
        #  Used for layout debugging.
        #
        if False:
            first_page = self._pageNumber == 1
            border_left = self.border_left(first_page)
            border_right = self.border_right(first_page)
            border_bottom = self.border_bottom(first_page)
            border_top = self.border_top(first_page)
            self.setStrokeColor(colors.red)
            self.line(border_left, border_bottom, border_right, border_bottom)  # bottom
            self.line(border_right, border_bottom, border_right, border_top)  # right
            self.line(border_left, border_bottom, border_left, border_top)  # left
            self.line(border_left, border_top, border_right, border_top)  # top
            self.setStrokeColor(colors.black)

    def showPage(self):
        self._saved_page_states.append(dict(self.__dict__))
        self._startPage()

        self.draw_remaining_page_background()
        self.draw_footer_image_block()

    def save(self):

        """
        Add page info to each page. Exactly what gets added will depend on which mixins
        have been added to the derived class.
        """

        num_pages = len(self._saved_page_states)

        for state in self._saved_page_states:
            self.__dict__.update(state)
            if self.status_method is not None:
                self.status_method(f'Processing page {self._pageNumber}/{num_pages}')
            self.draw_footer_image_block()
            self.draw_page_number(num_pages)
            canvas.Canvas.showPage(self)

        canvas.Canvas.save(self)

    def draw_background(self, image):
        if image is not None:
            self.drawImage(ImageReader(image), 0, 0, self.page_width(), self.page_height())

    def draw_first_page_background(self):
        """
        Adds the first page background image to the printout
        """
        self.draw_background(self.background_image_first)

    def draw_remaining_page_background(self):
        """
        Adds the background image for the remaining pages in the printout
        """
        self.draw_background(self.background_image_remaining)

    def draw_footer_image_block(self):
        """
        Adds the footer image to the printout
        """
        if self.background_image_footer is not None:
            border_bottom = self.border_bottom(True)
            page_width = self.page_width()
            self.drawImage(ImageReader(self.background_image_footer), 0, 0, page_width, border_bottom)



import uuid

from ajax_helpers.utils import ajax_command
from ajax_helpers.websockets.tasks import TaskHelper
from django.core.cache import caches
from django.urls import reverse

from django_advanced_pdf.engine.report_xml import ReportXML


class TaskProcessPDFHelper(TaskHelper):
    cache_key = 'default'

    def update_progress(self, message):
        self.update_state(state='PROGRESS', meta={'message': message})

    def get_xml_and_filename(self, slug):
        xml = '''
                    <document title="My test printout" page_size="A4">
                    <table><tr><td>Hello world</td></tr></table>
                    </document>
                '''
        return xml, 'hello world.pdf'

    def build_pdf(self, slug):
        xml, filename = self.get_xml_and_filename(slug)
        report_xml = ReportXML(status_method=self.update_progress)
        result = report_xml.load_xml_and_make_pdf(xml)
        return result, filename

    def get_config(self):
        return {'progress': False, 'message': 'initial', 'title': 'Processing....'}

    def pre_process(self, slug, **kwargs):
        pass  # can load user / set tenant here etc

    def process(self, config=False, slug=None, **kwargs):
        if config:
            return self.get_config()

        self.pre_process(slug=slug, **kwargs)
        result, filename = self.build_pdf(slug)
        hash_key = str(uuid.uuid4().hex)
        caches[self.cache_key].set(hash_key, {'file': result.getvalue(), 'filename': filename})
        return self.post_process(file_key=hash_key, slug=slug, **kwargs)

    def post_process(self, file_key, slug, **kwargs):
        # noinspection PyNoneFunctionAssignment
        urlconf = self.get_urlconf()
        if slug is not None and slug.get('modal') == '1':
            url = reverse('django_advanced_pdf:view_task_pdf_modal',
                          kwargs={'slug': f'file_key-{file_key}'},
                          urlconf=urlconf)
            return {'commands': [ajax_command('close'),
                                 ajax_command('show_modal', modal=url)]}
        else:
            url = reverse('django_advanced_pdf:view_task_pdf',
                          kwargs={'file_key': file_key},
                          urlconf=urlconf)
            return {'commands': [ajax_command('close'),
                                 ajax_command('redirect', url=url)]}

    def get_urlconf(self):
        return None



from django.apps import AppConfig


class DjangoAdvancedPdfConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'django_advanced_pdf'
    verbose_name = 'Django advanced pdfs'



class EnhancedTableData(object):
    """
    A helper class for passing data to the EnhancedTable class.
    """

    def __init__(self, row_data=None, commands=None, row_heights=None):
        """
        Class Constructor.

        @type   row_data : list of lists
        @param  row_data : A list of table row data
        @type   commands : list
        @param  commands : a list of command / styles for use with the row_data
        @type   row_heights : list
        @param  row_heights : a list of row heights as expected by ReportLab Table
        """
        self.row_data = []
        self.row_variables = []
        self.row_properties = []
        self.commands = []
        self.row_heights = []
        if row_data and isinstance(row_data, list):
            self.row_data = row_data
        if commands and isinstance(commands, list):
            self.commands = commands
        if row_heights and isinstance(row_heights, list):
            self.row_heights = row_heights
        self.row_length = len(self.row_data)
        self.rows_height = sum(self.row_heights)

    def reset(self):
        self.row_data = []
        self.row_variables = []
        self.row_properties = []
        self.commands = []
        self.row_length = 0
        self.rows_height = 0



from time import strftime, localtime

from reportlab.lib.units import mm
from reportlab.pdfbase.pdfmetrics import stringWidth

from django_advanced_pdf.engine.utils import GREYFILL_COLOUR, BLACK_COLOUR
from django_advanced_pdf.pagers.base import BasePager


class BorderPager(BasePager):
    def __init__(self, heading, filename, footer_field=None, pagesize=None, bottom_up=1, page_compression=None,
                 invariant=None, verbosity=0, encrypt=None, crop_marks=None, pdf_version=None, enforce_color_space=None,
                 bottom_text=None, test_mode=False, status_method=None, **kwargs):
        BasePager.__init__(self, heading=heading, filename=filename, pagesize=pagesize,
                           bottom_up=bottom_up, page_compression=page_compression,
                           invariant=invariant, verbosity=verbosity, encryp=encrypt,
                           crop_marks=crop_marks, pdf_version=pdf_version,
                           enforce_color_space=enforce_color_space, test_mode=test_mode,
                           status_method=status_method,
                           **kwargs)

        self._saved_page_states = []
        self.bottom_text = bottom_text

        self.set_page_used(top=self.page_height() - (self.border_top() - 5 * mm),
                           bottom=(self.border_bottom() + 7 * mm),
                           left=self.border_left(),
                           right=self.page_width() - (self.border_right()))
        self.footer_field = footer_field

    def draw_borders(self, page_count):
        self.setFont("Helvetica", 7)

        border_left = self.border_left()
        border_right = self.border_right()
        border_top = self.border_top()
        border_bottom = self.border_bottom()

        bottom_base = border_bottom - 7 * mm
        bottom_centre = border_bottom - 5 * mm

        self.setStrokeColor(GREYFILL_COLOUR)
        self.setFillColor(GREYFILL_COLOUR)

        text_width = stringWidth(self.heading, "Helvetica", 14)
        if text_width * mm < 30 * mm:
            text_width = 30 * mm
        box_left = (border_right - 5 * mm) - text_width

        self.rect(border_left, border_top - 5 * mm, box_left - border_left, 10 * mm, fill=1)
        self.setStrokeColor(BLACK_COLOUR)
        self.setFillColor(BLACK_COLOUR)
        self.line(border_left, bottom_base, border_right, bottom_base)  # bottom
        self.line(border_right, bottom_base, border_right, border_top + 5 * mm)  # right
        self.line(border_left, bottom_base, border_left, border_top + 5 * mm)  # left
        self.line(border_left, border_top + 5 * mm, border_right, border_top + 5 * mm)  # top
        self.line(border_left, border_top - 5 * mm, border_right, border_top - 5 * mm)  # top lower

        # the left point of the top box

        self.line(box_left, border_top + 5 * mm, box_left, border_top - 5 * mm)  # left
        mid_box = (border_right + 2 * mm - box_left) / 2
        self.setFont("Helvetica", 14)
        self.drawCentredString(border_right - mid_box, border_top - 2 * mm, self.heading)

        # page totals

        of_width = stringWidth(" of ", "Helvetica", 12)
        current_page_width = stringWidth("%d" % self._pageNumber, "Helvetica-Bold", 14)
        total_page_number_width = stringWidth("%d" % page_count, "Helvetica-Bold", 14)

        total_width = current_page_width + of_width + total_page_number_width + (2 * mm)

        mid_point = (of_width / 2) + total_page_number_width + (1 * mm)

        self.setFont("Helvetica", 10)
        self.drawCentredString(border_right - mid_point, bottom_centre, " of ")

        self.setFont("Helvetica-Bold", 12)
        self.drawRightString(border_right - (of_width / 2) - mid_point, bottom_centre, "%d" % self._pageNumber)
        self.drawString(border_right - mid_point + (of_width / 2), bottom_centre, "%d" % page_count)

        self.setFont("Helvetica", 10)

        time_str = self.get_time_string()
        time_width = stringWidth(time_str, "Helvetica", 10) + 4 * mm

        self.line(border_right - total_width, border_bottom, border_right - total_width, bottom_base)  # mid line
        self.line(border_right - (total_width + time_width), border_bottom,
                  border_right - (total_width + time_width), bottom_base)  # left line

        self.line(border_left, border_bottom, border_right, border_bottom)  # top line

        self.drawCentredString(border_right - (total_width + (time_width / 2)), bottom_centre, time_str)

        self.setFont("Helvetica-Bold", 12)
        if self.footer_field is not None:
            self.drawString(border_left + 2 * mm, border_top - 2 * mm, self.footer_field)
        # end page totals

        if self.bottom_text is not None:
            self.setFont("Helvetica", 10)
            self.drawString(border_left + 2 * mm, bottom_centre, self.bottom_text)

        self.display_program_name()

    def get_time_string(self):

        if self.test_mode:
            return "12:28 23-02-23"

        return strftime("%H:%M %d-%m-%y", localtime())


import logging

import reportlab
from reportlab.lib import colors
from reportlab.lib.colors import HexColor
from reportlab.lib.pagesizes import A4, A6, A5, A3, A2, A1, A0, LETTER, LEGAL, ELEVENSEVENTEEN, landscape, portrait
from reportlab.lib.units import mm
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.pdfgen import canvas
from reportlab.platypus import BaseDocTemplate, Frame, PageTemplate, SimpleDocTemplate
from django.contrib.humanize.templatetags.humanize import intcomma

logger = logging.getLogger("reportlab.platypus")

GREYFILL_COLOUR = HexColor(0xe3e3e3)
PURPLE_COLOUR = HexColor(0x721472)
BLACK_COLOUR = HexColor(0x000000)


class DocTemplate(SimpleDocTemplate):
    def __init__(self, heading, pager, *args, **kwargs):
        BaseDocTemplate.__init__(self, *args, **kwargs)
        #  Create and add two page templates each comprising a single frame to handle the
        #  first and subsequent pages. Each frame spans the entire page, but has padding
        #  set corresponding to the page used by the boxes and whatnot.

        frame1 = Frame(0,
                       0,
                       self.pagesize[0],
                       self.pagesize[1],
                       pager.page_used_left(),
                       pager.page_used_bottom() + 5 * mm,
                       pager.page_used_right(),
                       pager.page_used_top() + 5 * mm)
        frame2 = Frame(0,
                       0,
                       self.pagesize[0],
                       self.pagesize[1],
                       pager.page_used_left(False),
                       pager.page_used_bottom(False) + 5 * mm,
                       pager.page_used_right(False),
                       pager.page_used_top(False) + 5 * mm)

        self.addPageTemplates([PageTemplate(id='Page1', frames=frame1), PageTemplate(id='Page2', frames=frame2)])

        self.pager = pager
        self.heading = heading

    def handle_pageEnd(self):
        #  The _nextPageTemplateIndex setting is cleared on each page so we need to
        #  force it back to one so all pages after the first get the second page
        #  template.
        self._nextPageTemplateIndex = 1
        BaseDocTemplate.handle_pageEnd(self)

    def _startBuild(self, filename=None, canvasmaker=canvas.Canvas):
        self._calc()

        #  Each distinct pass gets a sequencer
        self.seq = reportlab.lib.sequencer.Sequencer()

        self.canv = canvasmaker(
            self.heading,
            filename or self.filename,
            pagesize=self.pagesize,
            invariant=self.invariant,
            page_compression=self.pageCompression,
            enforce_color_space=self.enforceColorSpace)

        getattr(self.canv, 'setEncrypt', lambda x: None)(self.encrypt)

        self.canv._cropMarks = self.cropMarks
        self.canv.setAuthor(self.author)
        self.canv.setTitle(self.title)
        self.canv.setSubject(self.subject)
        self.canv.setCreator(self.creator)
        self.canv.setKeywords(self.keywords)

        if self._onPage:
            self.canv.setPageCallBack(self._onPage)

        self.handle_documentBegin()

    def handle_pageBegin(self):
        """Perform actions required at beginning of page.
        shouldn't normally be called directly"""
        self.page += 1
        if self._debug:
            logger.debug("beginning page %d" % self.page)

        self.before_draw_canvas(self.canv, self.page)
        self.pageTemplate.beforeDrawPage(self.canv, self)
        self.pageTemplate.checkPageSize(self.canv, self)
        self.pageTemplate.onPage(self.canv, self)
        for f in self.pageTemplate.frames:
            f._reset()
        self.beforePage()
        # keep a count of flowables added to this page.  zero indicates bad stuff
        self._curPageFlowableCount = 0
        if hasattr(self, '_nextFrameIndex'):
            del self._nextFrameIndex
        self.frame = self.pageTemplate.frames[0]
        self.frame._debug = self._debug
        self.handle_frameBegin()

    def before_draw_canvas(self, canv, page_number):
        # if page_number == 2:
        #     self.create_abs_table(canv)
        pass
    # def create_abs_table(self, canv):
    #     data = [['00', '01', '02', '03', '04'],
    #             ['10', '11', '12', '13', '14'],
    #             ['20', '21', '22', '23', '24'],
    #             ['30', '31', '32', '33', '34']]
    #
    #     t = Table(data)
    #     t.setStyle(TableStyle([('BACKGROUND', (1, 1), (-2, -2), colors.green),
    #                            ('TEXTCOLOR', (0, 0), (1, -1), colors.red)]))
    #
    #     t.wrapOn(canv, self.width - 100, self.height)
    #     t.drawOn(canv, *self.coord(100, 200, mm))

    def coord(self, x, y, unit=mm):
        """
        http://stackoverflow.com/questions/4726011/wrap-text-in-a-table-reportlab
        Helper class to help position flowables in Canvas objects
        """
        x, y = x * unit, self.height - y * unit
        return x, y




class Grouper(dict):
    """
    Ordered grouping support class. When a key/value pair is added then the value
    is appended to a list, which can later be retrieved to get the values back in
    insertion order. Note that this does not support deletion or key replacement.

    The "group" method takes a list of values on which grouping should be based.
    These are concatenated as strings, except that floating values are chopped
    at three decimal places.

    """
    def __init__(self):
        self._ordered = []

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        self._ordered.append((key, value))

    def ordered(self):
        return [o[1] for o in self._ordered]

    def ordered_with_keys(self):
        return self._ordered

    @staticmethod
    def group(*args):
        bits = []
        for a in args:
            if isinstance(a, float):
                a = int(a * 1000)
            bits.append(str(a))
        return ','.join(bits)


class DecimalText(object):
        """
        Helper class to get decimal alignment in tables but with access
        to the common style settings. Note that this only works when used
        with table rows added using the L{append_table_data} method, since that automatically appends
        the relevant table style commands to make it work.
        """
        cachedWidths = {}

        def __init__(self, text, style, dp, stripped=False):
            self.m_text = text
            self.m_style = style
            self.m_dp = dp
            self.strip_trailing_zeros = stripped
            if self.strip_trailing_zeros:
                self.m_text = self.m_text.rstrip("0").rstrip(".")

        def font_name(self):
            return self.m_style.fontName

        def font_size(self):
            return self.m_style.fontSize

        def leading(self):
            return self.m_style.leading

        def right_padding(self):
            key = '%s_%s_%s' % (self.m_style.fontName, self.m_style.fontSize, self.m_dp)
            if key not in self.cachedWidths:
                if self.m_dp == 0:
                    w = 1 * mm
                else:
                    w = 1 * mm + stringWidth('.' + '0' * self.m_dp, self.m_style.fontName, self.m_style.fontSize)
                self.cachedWidths[key] = w
            return self.cachedWidths[key]

        def merge_variables(self, variables=None):
            if variables:
                self.m_text = self.m_text % variables
                if self.strip_trailing_zeros:
                    self.m_text = self.m_text.rstrip("0").rstrip(".")
            return self

        def __str__(self):
            return self.m_text


def get_page_size_from_string(page_size_string, landscape_portrait_string):
    page_size_string = page_size_string.lower()
    landscape_portrait_string = landscape_portrait_string.lower()
    if page_size_string == "a4":  # most common that why it at the top
        page_size = A4
    elif page_size_string == "a6":
        page_size = A6
    elif page_size_string == "a5":
        page_size = A5
    elif page_size_string == "a3":
        page_size = A3
    elif page_size_string == "a2":
        page_size = A2
    elif page_size_string == "a1":
        page_size = A1
    elif page_size_string == "a0":
        page_size = A0
    elif page_size_string == "leter":
        page_size = LETTER
    elif page_size_string == "legal":
        page_size = LEGAL
    elif page_size_string == "elevenseventeen":
        page_size = ELEVENSEVENTEEN
    else:
        assert False, "Unknown page type"

    if landscape_portrait_string == "landscape":
        return landscape(page_size)
    else:
        return portrait(page_size)


class PageUsed:
    def __init__(self, left, right, top, bottom):
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom


def intcomma_currency(value, show_more_decimal_places=False):

    if show_more_decimal_places:
        display_number = "%.5f" % float(value)
        main_number = display_number[:-3]
        extra_numbers = display_number[-3:]
        extra_numbers = extra_numbers.rstrip("0")
        value = intcomma("%.2f" % float(main_number))
        if extra_numbers != '':
            value += '<span class="currency_extra_digits">%s</span>' % extra_numbers
    else:
        value = intcomma("%.2f" % float(value))
    return value


class ColumnWidthPercentage(object):
    def __init__(self, value):
        self.value = value

    def get_value(self):
        return self.value


from reportlab.lib.colors import HexColor
from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import mm


class EnhancedParagraphStyle(ParagraphStyle):

    def process_css_for_table_paragraph_style(self, css, row_count, col_count):
        for style in css:
            if row_count is not None and col_count is not None and \
                    not self.is_valid_css_row(row_count, col_count, style[1], style[2]):
                continue
            style_type = style[0].lower()

            if style_type == '' or len(style) < 4:
                continue
            style_detail = style[3]
            if style_type in ('font_name',
                              'face',
                              'font'):
                self.fontName = style_detail
            elif style_type in ('size',
                                'fontsize'):
                self.fontSize = style_detail
            elif style_type == 'leading':
                self.leading = style_detail
            elif style_type == 'textcolor':
                self.textColor = style_detail
            elif style_type == 'align':
                alignment_type = style_detail.lower()
                if alignment_type == 'left':
                    self.alignment = TA_LEFT
                elif alignment_type == 'center':
                    self.alignment = TA_CENTER
                elif alignment_type == 'right':
                    self.alignment = TA_RIGHT

    def process_raw_css(self, css):
        styles_list = css.split(';')

        for style in styles_list:
            if style == '':
                continue
            style_type, style_detail = style.split(':')
            style_type = style_type.lower().lstrip("\r\n ")

            if style_type == '':
                continue
            if style_type in ('font_name',
                              'face',
                              'font'):
                self.fontName = style_detail
            elif style_type in ('size',
                                'font_size'):
                self.fontSize = int(style_detail)
            elif style_type == 'leading':
                self.leading = int(style_detail)
            elif style_type == 'text_color':
                self.textColor = HexColor(style_detail)
            elif style_type == 'back_color':
                self.backColor = HexColor(style_detail)
            elif style_type == 'align':
                alignment_type = style_detail.lower()
                if alignment_type == 'left':
                    self.alignment = TA_LEFT
                elif alignment_type == 'center':
                    self.alignment = TA_CENTER
                elif alignment_type == 'right':
                    self.alignment = TA_RIGHT
            elif style_type == 'left_indent':
                self.leftIndent = float(style_detail) * mm
            elif style_type == 'right_indent':
                self.rightIndent = float(style_detail) * mm
            elif style_type == 'first_line_indent':
                self.firstLineIndent = float(style_detail) * mm
            elif style_type == 'space_before':
                self.spaceBefore = float(style_detail) * mm
            elif style_type == 'space_after':
                self.spaceAfter = float(style_detail) * mm
            elif style_type == 'border_width':
                self.border_width = int(style_detail)
            elif style_type == 'border_padding':
                self.borderPadding = float(style_detail) * mm
            elif style_type == 'border_color':
                self.borderColor = HexColor(style_detail)
            elif style_type == 'border_radius':
                self.borderRadius = float(style_detail)

    @staticmethod
    def is_valid_css_row(row_count, col_count, css_start, css_end):
        if css_start == (0, 0) and css_end == (-1, -1):
            return True
        if css_start[1] <= row_count <= css_end[1]:
            if css_start[0] == 0 and css_end[0] == -1:
                return True
            if css_start[0] <= col_count <= css_end[0]:
                return True
        return False



from django.contrib import admin

from django_advanced_pdf.models import PrintingTemplate


@admin.register(PrintingTemplate)
class PrintingTemplateAdmin(admin.ModelAdmin):
    list_display = ('name',
                    )



from django.core.cache import caches
from django.core.files.base import ContentFile
from django.http import FileResponse, Http404
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.safestring import mark_safe
from django.views import View
from django.views.decorators.clickjacking import xframe_options_exempt
from django_modals.modals import Modal


@method_decorator(xframe_options_exempt, name='dispatch')
class ViewTaskPDF(View):
    cache_key = 'default'

    def get(self, request, file_key):
        cache_data = caches[self.cache_key].get(file_key)
        if cache_data is not None:
            response = FileResponse(ContentFile(cache_data['file']), content_type="application/pdf")
            filename = cache_data['filename']
            response['Content-Disposition'] = f'filename={filename}'
            response['Content-Length'] = len(cache_data['file'])
            caches[self.cache_key].delete(file_key)
            return response
        else:
            raise Http404


class PrintReportModal(Modal):
    button_container_class = 'text-center'
    size = 'xl'

    def get_url(self):
        file_key = self.slug.get('file_key')
        return reverse('django_advanced_pdf:view_task_pdf', kwargs={'file_key': file_key})

    def modal_content(self):
        url = self.get_url()
        html = mark_safe(f"""<iframe src="{url}" style="height:80vh; width:100%"></iframe>""")
        return html



from reportlab.platypus import Paragraph
from reportlab.platypus.paragraph import textTransformFrags

from django_advanced_pdf.engine.enhanced_paragraph.parser import EnhancedParaParser


class EnhancedParagraph(Paragraph):
    # bulletText needs to be camelcase as it is referenced internally by the reportlab code and will break if changed
    def __init__(self, text, style, bulletText=None, frags=None, case_sensitive=1,
                 encoding='utf8', css_classes=None):
        if css_classes is None:
            self.css_classes = {}
        else:
            self.css_classes = css_classes
        Paragraph.__init__(self, text, style, bulletText, frags, case_sensitive, encoding)

    def _setup(self, text, style, bullet_text, frags, cleaner):
        if frags is None:
            text = cleaner(text)
            _parser = EnhancedParaParser(self.css_classes)
            _parser.caseSensitive = self.caseSensitive
            style, frags, bullet_text_frags = _parser.parse(text, style)
            if frags is None:
                raise ValueError("xml parser error (%s) in paragraph beginning\n'%s'"
                                 % (_parser.errors[0], text[:min(30, len(text))]))
            textTransformFrags(frags, style)
            if bullet_text_frags:
                bullet_text = bullet_text_frags

        # AR hack
        self.text = text
        self.frags = frags
        self.style = style
        self.bulletText = bullet_text
        self.debug = 0

    def calc_text_height(self, avail_width):
        # work out widths array for breaking
        style = self.style
        left_indent = style.leftIndent
        first_line_width = avail_width - (left_indent + style.firstLineIndent) - style.rightIndent
        later_widths = avail_width - left_indent - style.rightIndent
        _wrapWidths = [first_line_width, later_widths]
        if style.wordWrap == 'CJK':
            # use Asian text wrap algorithm to break characters
            bl_para = self.breakLinesCJK(_wrapWidths)
        else:
            bl_para = self.breakLines(_wrapWidths)
        auto_leading = getattr(self, 'autoLeading', getattr(style, 'autoLeading', ''))
        leading = style.leading
        if bl_para.kind == 1:
            if auto_leading not in ('', 'off'):
                height = 0
                if auto_leading == 'max':
                    for l in bl_para.lines:
                        height += max(l.ascent - l.descent, leading)
                elif auto_leading == 'min':
                    for l in bl_para.lines:
                        height += l.ascent - l.descent
                else:
                    raise ValueError('invalid autoLeading value %r' % auto_leading)
            else:
                height = len(bl_para.lines) * leading
        else:
            if auto_leading == 'max':
                leading = max(leading, bl_para.ascent - bl_para.descent)
            elif auto_leading == 'min':
                leading = bl_para.ascent - bl_para.descent
            height = len(bl_para.lines) * leading
        return height



from django.http import HttpResponse
from django.views.generic import DetailView

from django_advanced_pdf.models import PrintingTemplate


class DatabasePDFView(DetailView):
    model = PrintingTemplate

    def get_pager_kwargs(self):
        return {}

    def render_to_response(self, context, **kwargs):
        result = self.object.make_pdf(context=context, **self.get_pager_kwargs())
        response = HttpResponse(result.getvalue(), content_type='application/pdf')
        return response




import base64
import pickle

from reportlab.lib.units import mm
from reportlab.platypus import Image
from io import BytesIO  # for Python 3


def insert_image(tag):
    image_data = BytesIO(base64.b64decode(tag.text))
    image = Image(image_data)
    if tag.get('width') is not None:
        image.drawWidth = float(tag.get('width')) * mm
    if tag.get('height') is not None:
        image.drawHeight = float(tag.get('height')) * mm
    return image


def insert_obj(tag):
    data = tag.text
    if data is not None:
        enc_data = data.encode('ascii')
        object_data = base64.urlsafe_b64decode(enc_data)
        report_object = pickle.loads(object_data)
        return report_object
    return None



from reportlab import rl_config
from reportlab.lib.rl_accel import fp_str
from reportlab.lib.utils import annotateException, flatten
from reportlab.platypus.flowables import Flowable
from reportlab.platypus.flowables import PageBreak
from reportlab.platypus.para import handleSpecialCharacters
from reportlab.platypus.paragraph import Paragraph
# noinspection PyProtectedMember
from reportlab.platypus.tables import Table, _calc_pc, spanFixDim, CellStyle
from six import string_types

from django_advanced_pdf.engine.enhanced_table.data_paragraph import DataParagraph
from django_advanced_pdf.engine.utils import DecimalText

OVERFLOW_ROW = -9999
HEADER_FOOTER = 7777
UNDEFINED_ROW = -8888

KEEP_TYPE_NA = 0
KEEP_TYPE_SPAN = 1
KEEP_TYPE_START = 2
KEEP_TYPE_MIDDLE = 3
KEEP_TYPE_END = 4


# noinspection PyPep8Naming
class EnhancedTable(Table):
    """
    An enhanced version of Reportlab's Table class.
    Key differences are the ability to supply details for continuation header/footer which get used when a
    table splits across page boundaries.
    Data rows can be supplied with corresponding properties which indicate if a particular row is a header, total data
    or blank row. This information is used to help fine-tune where a table can be split should the need arise.
    """

    def __init__(self, table_data, headers=None, footers=None, min_rows_after_header=1, min_rows_before_total=1,
                 col_widths=None, row_heights=None, style=None,
                 repeat_rows=0, repeat_cols=0, split_by_row=1, empty_table_action=None, ident=None,
                 h_align=None, v_align=None, normalized_data=0, cell_styles=None,
                 _calc_row_splits=True, initial=False, pos_x=None, pos_y=None, colpositions=None):
        """
        Class Constructor.

        Custom parameters
        =================
        @type   table_data : dict
        @param  table_data : dictionary of row data (key='row_data'), variables ('variables') and
                             properties ('properties') for each row of this table (keys are 'row_data', 'variables'
                             and 'properties' respectively). Each contains a list of data, one entry per row.
                             Variables contains the substitution values for continuation headers and footers.
                             Properties are used primarily to indicate whether a row is a header, total, data or blank.
        @type   headers : list
        @param  headers : list of row data and styles for a continuation header, if required
        @type   footers : list
        @param  footers : list of row row data and styles for a continuation footer, if required
        @type   min_rows_after_header : int
        @param  min_rows_after_header : The number of rows to mark as non-splittable after any row marked with a
                                        property of 'HEADER'
        @type   min_rows_before_total : int
        @param  min_rows_before_total : The number of rows to mark as non-splittable immediately prior to any row
                                        marked with a property of 'TOTAL'
        @type   _calc_row_splits      : bool
        @param  _calc_row_splits      : private parameter used to control the calculation of nosplit commands

        Parameters inherited from Reportlab's reportlab.platypus.tables.Table class
        =================
        @type   col_widths : list
        @param  col_widths : a list of column widths as expected by ReportLab Table
        @type   row_heights : list
        @param  row_heights : a list of row heights as expected by ReportLab Table
        @type   style : list
        @param  style : a list of row styles (commands) as expected by ReportLab Table
        @type   repeat_rows : int
        @param  repeat_rows : The number of initial rows to repeat on subsequent pages if a table is split vertically
        @type   repeat_cols : int
        @param  repeat_cols : NOT SUPPORTED (ReportLab functionailty)
        @type   h_align     : str
        @param  h_align     : horizontal alignment as expected by ReportLab Table
        @type   v_align     : str
        @param  v_align     : vertical alignment as expected by ReportLab Table
        @type   normalized_data     : int
        @param  normalized_data     : Flag to indictae whether teh table data passed in is already normalised or not
        @type   cell_styles : list
        @param  style : a list of cell styles (commands) as expected by ReportLab Table
        @rtype  : EnhancedTable
        @return : An Enhanced Table object, based on ReportLab Table
        """
        self._height = 0
        self.availWidth = 0
        self.initial = initial
        self.headers = headers
        self.footers = footers
        self.data = table_data.get('row_data', [])
        self.variables = table_data.get('row_variables', [{} for _ in range(len(self.data))])
        self.properties = table_data.get('row_properties', [{} for _ in range(len(self.data))])
        self.keep_with_next = table_data.get('keep_with_next', [KEEP_TYPE_NA for _ in range(len(self.data))])
        self.headers_index = table_data.get('headers_index', [False for _ in range(len(self.data))])
        self.footers_index = table_data.get('footers_index', [False for _ in range(len(self.data))])
        self.min_rows_after_header = min_rows_after_header
        self.min_rows_before_total = min_rows_before_total
        self.pos_x = pos_x
        self.pos_y = pos_y
        self._colpositions = colpositions

        no_split_cmds = self._calc_nosplit_positions(_calc_row_splits)

        if no_split_cmds:
            if style:
                style += no_split_cmds
            else:
                style = no_split_cmds

        Table.__init__(self,
                       data=self.data,
                       colWidths=col_widths,
                       rowHeights=row_heights,
                       style=style,
                       repeatRows=repeat_rows,
                       repeatCols=repeat_cols,
                       splitByRow=split_by_row,
                       emptyTableAction=empty_table_action,
                       ident=ident,
                       hAlign=h_align,
                       vAlign=v_align,
                       normalizedData=normalized_data,
                       cellStyles=cell_styles)

    def _getFirstPossibleSplitRowPosition(self, availHeight, ignoreSpans=0):
        # Note - this is actually looking for the BEST available split position, which is not necessarily the first.
        impossible = {}
        if self._spanCmds:
            self._getRowImpossible(impossible, self._rowSpanCells, self._spanRanges)
        if self._nosplitCmds:
            self._getRowImpossible(impossible, self._rowNoSplitCells, self._nosplitRanges)

        h = 0
        n = 1

        header_index = None
        footer_index = None

        split_at = 0  # from this point of view 0 is the first position where the table may *always* be split
        use_middle = True
        for i, (rh, header_index, footer_index, keep_with_next) in enumerate(zip(self._rowHeights,
                                                                                 self.headers_index,
                                                                                 self.footers_index,
                                                                                 self.keep_with_next)):

            number_of_header = 0
            if header_index is not None:
                number_of_header = self.headers[header_index].row_length

            footer_height = 0
            if footer_index is not None:
                footer_height = self.footers[footer_index].rows_height

            if h + rh > availHeight - footer_height:
                break

            if (self.initial or i > number_of_header) and n not in impossible and\
                    keep_with_next in [KEEP_TYPE_NA, KEEP_TYPE_END, KEEP_TYPE_MIDDLE]:

                if keep_with_next in [KEEP_TYPE_NA, KEEP_TYPE_END]:
                    use_middle = False
                    split_at = n
                elif use_middle:
                    split_at = n
            h = h + rh
            n += 1

        return split_at, header_index, footer_index

    @staticmethod
    def merge_variables_into_data(data, variables):
        """
        This takes a list of table data and combines it with its associated variables (if supplied)
        This allows us to substitute variables into rows to create header/footer rows for use when splitting tables.
        """

        output = []
        for row in data:
            row_data = []
            for col in row:
                if isinstance(col, DataParagraph):
                    row_data.append(col.merge_variables(variables))
                elif isinstance(col, Paragraph):
                    row_data.append(col)
                    # row_data.append(col.merge_variables(variables))
                elif isinstance(col, DecimalText):
                    row_data.append(col.merge_variables(variables))
                else:
                    if col != '':
                        try:
                            value = ""
                            col = col.decode('utf-8')
                            col = handleSpecialCharacters(None, col)
                            col = ''.join(col).encode('utf-8', 'ignore')
                            for c in col:
                                value += c
                            col = value
                        except Exception:
                            pass
                    if col and variables:
                        try:
                            row_data.append(col % variables)
                        except TypeError:
                            # This is a Decimal Text, text needs to be removed, updated and reinserted
                            text = col.m_text
                            updated_text = text % variables
                            col.m_text = updated_text
                            row_data.append(col)
                    else:
                        row_data.append(col)
            output.append(row_data)

        return output

    def _cr_1_1_enhanced(self, n, repeat_rows, header_rows, cmds):
        # Modified version of Table._cr_1_1
        for c in cmds:
            c = tuple(c)
            (sc, sr), (ec, er) = c[1:3]
            if sr in ('splitfirst', 'splitlast'):
                self._addCommand(c)
            else:
                # if sr >= 0 and sr >= repeat_rows and sr < n and er >= 0 and er < n:
                if 0 <= sr < n and sr >= repeat_rows and 0 <= er < n:
                    # Ignore commands which are before the split_point (n) and start after repeat_rows
                    continue
                if repeat_rows <= sr < n:
                    sr = repeat_rows + header_rows
                elif sr >= repeat_rows and sr >= n:
                    sr = sr + repeat_rows + header_rows - n
                if repeat_rows <= er < n:
                    er = repeat_rows + header_rows
                elif er >= repeat_rows and er >= n:
                    er = er + repeat_rows + header_rows - n
                self._addCommand((c[0],) + ((sc, sr), (ec, er)) + c[3:])

    def _add_offset_commands(self, n, cmds):
        for c in cmds:
            c = tuple(c)
            (sc, sr), (ec, er) = c[1:3]
            sr += n
            er += n
            self._addCommand((c[0],) + ((sc, sr), (ec, er)) + c[3:])

    # noinspection DuplicatedCode
    def _splitRows(self, availHeight, doInRowSplit=0):

        n, header_index, footer_index = self._getFirstPossibleSplitRowPosition(availHeight, ignoreSpans=doInRowSplit)
        if n <= self.repeatRows:
            return []
        lim = len(self._rowHeights)
        if n == lim:  # No splitting required
            return [self]

        r0_end = n

        # Check to see if the row we are splitting on is of type 'BLANK'.
        # If it is then we can actually ignore it - i.e. end the first split section (RO) one row earlier.
        # This will probably have adverse side effects if the row had commands associated with it

        insert_pagebreak = False
        try:
            if n > 0 and self.properties[n - 1]['row_type'].upper() == 'BLANK':
                insert_pagebreak = True  # Otherwise reportlab can squeeze in more stuff afterwards, which is bad
                r0_end = n - 1
        except KeyError:  # row_type not specified
            pass

        repeat_rows = self.repeatRows
        repeat_cols = self.repeatCols
        split_by_row = self.splitByRow
        data = self._cellvalues

        # we're going to split into two superRows
        ident = self.ident

        footer_row_data = []
        footer_row_heights = []
        footer_cell_styles = []
        footer_row_variables = []
        footer_row_properties = []
        footer_keep_with_next = []
        footer_commands = []

        if footer_index is not None:
            insert_pagebreak = True
            footer_data = self.footers[footer_index]
            try:
                footer_row_data = self.merge_variables_into_data(footer_data.row_data, self.variables[r0_end - 1])
            except (IndexError, KeyError):
                footer_row_data = footer_data.row_data

            footer_row_heights = footer_data.row_heights
            footer_row_variables = [{} for _ in footer_row_data]
            footer_keep_with_next = [0 for _ in footer_row_data]
            footer_row_properties = [{'row_type': 'HEADING', 'SPLITTABLE': False} for _ in footer_row_data]
            footer_row_data = self.normalizeData(footer_row_data)
            footer_commands = footer_data.commands
            for i in range(footer_data.row_length):
                ncols = len(footer_row_data[i])
                cellcols = []
                for j in range(ncols):
                    cellcols.append(CellStyle('header_footer'))
                footer_cell_styles.append(cellcols)

        r0_table_data = {
            'row_data': data[:r0_end] + footer_row_data,
            'row_variables': self.variables[:r0_end] + footer_row_variables,
            'row_properties': self.properties[:r0_end] + footer_row_properties,
            'keep_with_next': self.keep_with_next[:r0_end] + footer_keep_with_next,
        }

        r0 = EnhancedTable(r0_table_data,
                           col_widths=self._colWidths,
                           row_heights=self._argH[:r0_end] + footer_row_heights,
                           cell_styles=self._cellStyles[:r0_end] + footer_cell_styles,
                           repeat_rows=repeat_rows, repeat_cols=repeat_cols,
                           split_by_row=split_by_row, normalized_data=1,
                           ident=ident,
                           min_rows_after_header=self.min_rows_after_header,
                           min_rows_before_total=self.min_rows_before_total,
                           _calc_row_splits=False,
                           colpositions=self._colpositions)

        # copy the commands

        A = []
        # hack up the line commands
        for op, (sc, sr), (ec, er), weight, color, cap, dash, join, count, space in self._linecmds:

            if isinstance(sr, string_types) and sr.startswith('split'):
                A.append((op, (sc, sr), (ec, sr), weight, color, cap, dash, join, count, space))
                if sr == 'splitlast':
                    sr = er = n - 1
                elif sr == 'splitfirst':
                    sr = n
                    er = n

            if sc < 0:
                sc = sc + self._ncols
            if ec < 0:
                ec = ec + self._ncols
            if sr < 0:
                sr = sr + self._nrows
            if er < 0:
                er = er + self._nrows

            if op in ('BOX', 'OUTLINE', 'GRID'):
                if sr < n <= er:
                    # we have to split the BOX
                    A.append(('LINEABOVE', (sc, sr), (ec, sr), weight, color, cap, dash, join, count, space))
                    A.append(('LINEBEFORE', (sc, sr), (sc, er), weight, color, cap, dash, join, count, space))
                    A.append(('LINEAFTER', (ec, sr), (ec, er), weight, color, cap, dash, join, count, space))
                    A.append(('LINEBELOW', (sc, er), (ec, er), weight, color, cap, dash, join, count, space))
                    if op == 'GRID':
                        A.append(('LINEBELOW', (sc, n - 1), (ec, n - 1), weight, color, cap, dash, join, count, space))
                        A.append(('LINEABOVE', (sc, n), (ec, n), weight, color, cap, dash, join, count, space))
                        A.append(('INNERGRID', (sc, sr), (ec, er), weight, color, cap, dash, join, count, space))
                else:
                    A.append((op, (sc, sr), (ec, er), weight, color, cap, dash, join, count, space))
            elif op in ('INNERGRID', 'LINEABOVE'):
                if sr < n <= er:
                    A.append(('LINEBELOW', (sc, n - 1), (ec, n - 1), weight, color, cap, dash, join, count, space))
                    A.append(('LINEABOVE', (sc, n), (ec, n), weight, color, cap, dash, join, count, space))
                A.append((op, (sc, sr), (ec, er), weight, color, cap, dash, join, count, space))
            elif op == 'LINEBELOW':
                if sr < n and er >= (n - 1):
                    A.append(('LINEABOVE', (sc, n), (ec, n), weight, color, cap, dash, join, count, space))
                A.append((op, (sc, sr), (ec, er), weight, color))
            elif op == 'LINEABOVE':
                if sr <= n <= er:
                    A.append(('LINEBELOW', (sc, n - 1), (ec, n - 1), weight, color, cap, dash, join, count, space))
                A.append((op, (sc, sr), (ec, er), weight, color, cap, dash, join, count, space))
            else:
                A.append((op, (sc, sr), (ec, er), weight, color, cap, dash, join, count, space))

        # The following add back all the row commands (munged above) for the first n rows

        r0._cr_0(n, A, self._nrows)

        r0._cr_0(n, self._bkgrndcmds, self._nrows)
        r0._cr_0(n, self._spanCmds, self._nrows)
        r0._cr_0(n, self._nosplitCmds, self._nrows)
        # r0._cr_0_footer(n, footer_commands)
        r0._cr_1_0(HEADER_FOOTER-n, footer_commands)

        # Now we need to add any footer styles back on to the end (with all their cell ranges shifted)
        header_row_data = []
        header_row_heights = []
        header_row_styles = []
        header_cell_styles = []
        header_row_variables = []
        header_row_properties = []
        header_keep_with_next = []
        header_commands = []
        blank_header_data = []

        if header_index is not None:
            insert_pagebreak = True
            header_data = self.headers[header_index]
            try:
                header_row_data = self.merge_variables_into_data(header_data.row_data, self.variables[n - 1])
            except (IndexError, KeyError):
                # If there are no variables supplied (i.e. it's a static header)
                header_row_data = header_data.row_data
            header_row_heights = header_data.row_heights

            for i in range(header_data.row_length):
                ncols = len(header_row_data[i])
                cellcols = []
                for j in range(ncols):
                    cellcols.append(CellStyle('header_footer'))
                header_cell_styles.append(cellcols)
            header_row_variables = [{} for _ in header_row_data]
            header_keep_with_next = [False for _ in header_row_data]
            blank_header_data = [None for _ in header_row_data]
            header_row_properties = [{'row_type': 'HEADING', 'SPLITTABLE': False} for _ in header_row_data]
            header_row_data = self.normalizeData(header_row_data)
            header_commands = header_data.commands

        # Construct the R1 row data, heights and cell styles.
        # NB. this should work even if repeatRows is 0 (resulting in empty lists, which collapse to nothing)

        r1_table_data = {
            'row_data': data[:repeat_rows] + header_row_data + data[n:],
            'row_variables': ([{} for _ in range(0, len(data[:repeat_rows]))] +
                              header_row_variables + self.variables[n:]),
            'row_properties': ([{} for _ in range(0, len(data[:repeat_rows]))] +
                               header_row_properties + self.properties[n:]),
            'keep_with_next': ([{} for _ in range(0, len(data[:repeat_rows]))] +
                               header_keep_with_next + self.keep_with_next[n:]),
            'headers_index': ([None for _ in range(0, len(data[:repeat_rows]))] +
                              blank_header_data + self.headers_index[n:]),
            'footers_index': ([None for _ in range(0, len(data[:repeat_rows]))] +
                              blank_header_data + self.footers_index[n:]),
        }

        r1_row_heights = self._argH[:repeat_rows] + header_row_heights + self._argH[n:]
        r1_cell_styles = self._merge_cell_styles(first=self._cellStyles[:repeat_rows],
                                                 headers=header_cell_styles,
                                                 last=self._cellStyles[n:])
        r1 = EnhancedTable(r1_table_data,
                           col_widths=self._colWidths,
                           row_heights=r1_row_heights,
                           repeat_rows=repeat_rows, repeat_cols=repeat_cols,
                           split_by_row=split_by_row, normalized_data=1,
                           cell_styles=r1_cell_styles,
                           ident=ident,
                           headers=self.headers,
                           footers=self.footers,
                           min_rows_after_header=self.min_rows_after_header,
                           min_rows_before_total=self.min_rows_before_total,
                           _calc_row_splits=False,
                           colpositions=self._colpositions)

        # Need to account for any header rows added when we call the following otherwise the row
        # styles will get out of step

        header_rows = len(header_row_data)
        if repeat_rows > 0 or header_rows > 0:
            # the method _cr_1_1_enhaced moves all table row commands (styles) down by adjusting their ranges
            # It leaves styles affecting rows 0 - repeat_rows
            r1._cr_1_0(HEADER_FOOTER, header_commands)
            r1._cr_1_1_enhanced(n, repeat_rows, header_rows, A)
            r1._cr_1_1_enhanced(n, repeat_rows, header_rows, self._bkgrndcmds)
            r1._cr_1_1_enhanced(n, repeat_rows, header_rows, self._spanCmds)
            r1._cr_1_1_enhanced(n, repeat_rows, header_rows, self._nosplitCmds)
        else:
            # the method _cr_1_0 moves all line commands down up by n rows
            r1._cr_1_0(n - header_rows, A)
            r1._cr_1_0(n - header_rows, self._bkgrndcmds)
            r1._cr_1_0(n - header_rows, self._spanCmds)
            r1._cr_1_0(n - header_rows, self._nosplitCmds)

        # Now we need to add back in the styles for the header rows (leaving their start/end positions as-is).
        if header_rows:
            r1._add_offset_commands(repeat_rows, header_row_styles)

        r0.hAlign = r1.hAlign = self.hAlign
        r0.vAlign = r1.vAlign = self.vAlign
        self.onSplit(r0)
        self.onSplit(r1)

        if insert_pagebreak:
            return [r0, PageBreak(), r1]
        else:
            return [r0, r1]

    @staticmethod
    def _merge_cell_styles(first, headers, last):
        if len(headers) == 0:
            return first + last

        if len(first) > 0:
            col_len = len(first[0])
        elif len(last) > 0:
            col_len = len(last[0])
        else:
            col_len = 0

        headers_mod = []
        for row in headers:
            header_col_len = len(row)
            if header_col_len != col_len:
                for x in range(header_col_len, col_len):
                    row.append(CellStyle('header_footer'))

            headers_mod.append(row)

        return first + headers_mod + last

    def _calc_nosplit_positions(self, _calc_row_splits):
        no_split_cmds = []
        if _calc_row_splits is True:
            # Calculate which rows are splittable based on row types as (possibly) supplied in properties
            # Add appropriate NOSPLIT commands for each one

            num_rows = len(self.data)
            for i in range(num_rows):
                row_type = self.properties[i].get('row_type', 'data').upper()
                if self.properties[i].get('nosplit', False):
                    no_split_cmds.append(('NOSPLIT', (0, i), (-1, i + 1),))
                elif row_type == 'BLANK':
                    # Don't allow the table to split just before a blank row since this means that you get a blank
                    # as the first row of the next page which a) looks dodgy and b) is hard to remove in code
                    # since all the other row styles would need adjusting. Blank lines appearing as the last row are ok
                    # as they will be removed and replaced with PageBreak objects in _splitRows
                    no_split_cmds.append(('NOSPLIT', (0, i - 1), (-1, i)))
                elif row_type in ('HEADER', 'HEADING'):
                    no_split_cmds.append(('NOSPLIT', (0, i), (-1, i + self.min_rows_after_header)))
                elif row_type == 'TOTAL':
                    no_split_cmds.append(('NOSPLIT', (-1, i - self.min_rows_before_total), (0, i)))

        return no_split_cmds

    def drawOn(self, canvas, x, y, _sW=0):
        if self.pos_x is not None and self.pos_y is not None:
            x = float(self.pos_x)
            y = float(self.pos_y)

        return super(EnhancedTable, self).drawOn(canvas, x, y, _sW)

    def wrap(self, availWidth, availHeight):
        self._calc(availWidth, availHeight)
        self.availWidth = availWidth
        if self.pos_x is not None and self.pos_y is not None:
            self._height = 0
            self.availWidth = availWidth - float(self.pos_x)

        return self._width, self._height

    # noinspection PyProtectedMember
    def calc_height_of_table(self, availHeight, availWidth, H=None, W=None):
        _ = H  # remove pep8
        H = self._argH
        if not W:
            W = _calc_pc(self._argW, availWidth)  # widths array

        hmax = lim = len(H)
        longTable = self._longTableOptimize

        if None in H or OVERFLOW_ROW in H or UNDEFINED_ROW in H:
            canv = getattr(self, 'canv', None)
            saved = None
            # get a handy list of any cells which span rows. should be ignored for sizing
            if self._spanCmds:
                rowSpanCells = self._rowSpanCells
                colSpanCells = self._colSpanCells
                spanRanges = self._spanRanges
                colpositions = self._colpositions
            else:
                colpositions = None
                rowSpanCells = colSpanCells = ()
                spanRanges = {}
            if canv:
                saved = canv._fontname, canv._fontsize, canv._leading
            H0 = H
            H = H[:]  # make a copy as we'll change it
            self._rowHeights = H
            spanCons = {}

            find_types = []
            if None in H:
                find_types.append(None)
            if OVERFLOW_ROW in H:
                find_types.append(OVERFLOW_ROW)
            if UNDEFINED_ROW in H:
                find_types.append(UNDEFINED_ROW)

            while None in H or OVERFLOW_ROW in H or UNDEFINED_ROW in H:
                i = None
                next_find_type = None
                for find_type in find_types:
                    if find_type in H:
                        find_type_index = H.index(find_type)
                        if i is None or find_type_index < i:
                            i = find_type_index
                            next_find_type = find_type

                V = self._cellvalues[i]  # values for row i
                S = self._cellStyles[i]  # styles for row i
                h = 0

                for j, (v, s, w) in enumerate(list(zip(V, S, W))):  # value, style, width (lengths must match)
                    ji = j, i
                    if next_find_type == OVERFLOW_ROW:
                        s.leading = 1
                    span = spanRanges.get(ji, None)
                    if ji in rowSpanCells and not span:
                        continue  # don't count it, it's either occluded or unreliable
                    else:
                        if isinstance(v, (tuple, list, Flowable)):
                            if isinstance(v, Flowable):
                                v = (v,)
                            else:
                                v = flatten(v)
                            v = V[j] = self._cellListProcess(v, w, None)
                            if w is None and not self._canGetWidth(v):
                                raise ValueError("Flowable %s in cell(%d,%d) can't have auto width in\n%s" % (
                                    v[0].identity(30), i, j, self.identity(30)))
                            if canv:
                                canv._fontname, canv._fontsize, canv._leading = s.fontname, s.fontsize, s.leading or 1.2 * s.fontsize
                            if ji in colSpanCells:
                                if not span:
                                    continue
                                w = max(colpositions[span[2] + 1] - colpositions[span[0]], w)
                            dW, t = self._listCellGeom(v, w or self._listValueWidth(v), s)
                            if canv:
                                canv._fontname, canv._fontsize, canv._leading = saved
                            dW = dW + s.leftPadding + s.rightPadding
                            if not rl_config.allowTableBoundsErrors and dW > w:
                                from reportlab.platypus.doctemplate import LayoutError
                                raise LayoutError(
                                    "Flowable %s (%sx%s points) too wide for cell(%d,%d) (%sx* points) in\n%s" % (
                                        v[0].identity(30), fp_str(dW), fp_str(t), i, j, fp_str(w), self.identity(30)))
                        else:
                            v = (v is not None and str(v) or '').split("\n")
                            t = (s.leading or 1.2 * s.fontsize) * len(v)
                        t += s.bottomPadding + s.topPadding
                        if span:
                            r0 = span[1]
                            r1 = span[3]
                            if r0 != r1:
                                x = r0, r1
                                spanCons[x] = max(spanCons.get(x, t), t)
                                t = 0
                    if t > h:
                        h = t  # record a new maximum
                H[i] = h
                # we can stop if we have filled up all available room
                if longTable:
                    hmax = i + 1  # we computed H[i] so known len == i+1
                    height = int(sum(H[:hmax]) + 0.5)
                    i_avail_height = int(availHeight + 0.5)
                    if height > i_avail_height:
                        # we can terminate if all spans are complete in H[:hmax]
                        if spanCons:
                            msr = max(x[1] for x in spanCons.keys())  # RS=[endrowspan,.....]
                            if hmax > msr:
                                while None in H:
                                    next_none = H.index(None)
                                    H[next_none] = UNDEFINED_ROW
                                break
            if UNDEFINED_ROW not in H and OVERFLOW_ROW not in H:
                hmax = lim

            if spanCons:
                try:
                    spanFixDim(H0, H, spanCons)
                except:
                    annotateException(
                        '\nspanning problem in %s hmax=%s lim=%s avail=%s x %s\nH0=%r H=%r\nspanCons=%r' % (
                            self.identity(), hmax, lim, availWidth, availHeight, H0, H, spanCons))

        # iterate backwards through the heights to get rowpositions in reversed order
        self._rowpositions = j = []
        height = c = 0

        for i in range(hmax - 1, -1, -1):
            j.append(height)
            y = H[i] - c
            t = height + y
            c = (t - height) - y
            height = t
        j.append(height)
        j.reverse()  # reverse the reversed list of row positions
        return height, hmax

    def _calc_height(self, availHeight, availWidth, H=None, W=None):
        height, height_max = self.calc_height_of_table(availHeight, availWidth, H, W)
        self._height = height
        self._hmax = height_max

    def get_height(self):
        height, _ = self.calc_height_of_table(10000, 10000)
        return height

    def get_width(self):
        self._calc_width(10000)
        return self._width

    def spanFixDim(self, V0, V, spanCons, lim=None, FUZZ=rl_config._FUZZ):
        """
        Assigned the row height to row span fields. Report lab version equalised the spacing across all
        reports. We changed this to only do it on the last row of the col span
        :param V0:
        :param V:
        :param spanCons:
        :param lim:
        :param FUZZ:
        :return:
        """

        # assign required space to variable rows equally to existing calculated values
        M = {}
        if not lim:
            lim = len(V0)  # in longtables the row calcs may be truncated

        # we assign the largest spaces first hoping to get a smaller result
        for v, (x0, x1) in reversed(sorted(((iv, ik) for ik, iv in spanCons.items()))):
            if x0 >= lim:
                continue
            x1 += 1
            t = sum([V[x] + M.get(x, 0) for x in range(x0, x1-1)])

            if t >= v - FUZZ:
                continue  # already good enough

            X = [x for x in range(x0, x1) if V0[x] is None]  # variable candidates
            if not X:
                continue  # something wrong here mate

            M[X[-1]] = v - t

        for x, v in M.items():
            V[x] = v
